import abc
from functools import cached_property
from collections import defaultdict
from collections.abc import Collection, Iterable, Container, Sequence

import numpy as np
import pandas as pd

from localuf import constants
from localuf.noise.main import Noise
from localuf.type_aliases import Coord, Edge, EdgeType, Parametrization, NoiseModel, Node, DecodingScheme
from localuf._schemes import Scheme, Batch, Global, Forward, Frugal
from localuf.noise import CircuitLevel, CodeCapacity, Phenomenological


class Code(abc.ABC):
    """The decoding graph G = (V, E) of a CSS code."""

    _TIME_AXIS = -1
    _LONG_AXIS: int
    _CODE_DIMENSION: int

    def __init__(
            self,
            d: int,
            noise: NoiseModel,
            scheme: DecodingScheme = 'batch',
            window_height: int | None = None,
            commit_height: int | None = None,
            buffer_height: int | None = None,
            merge_equivalent_boundary_nodes: bool = False,
            parametrization: Parametrization = 'balanced',
            demolition: bool = False,
            monolingual: bool = False,
            _merge_redundant_edges: bool = True,
    ):
        """
        :param d: code distance.
        :param noise: noise model.
        :param scheme: the decoding scheme.
        :param window_height: total layer count in the time direction.
            Affects only batch and global batch decoding schemes.
            Default value is ``d``.
        :param commit_height: the layer count in the time direction
            that is committed.
            Affects only forward and frugal decoding schemes.
            Default value is ``d`` for forward scheme
            and ``1`` (``2*(d//2)``) for frugal scheme.
        :param buffer_height: the layer count in the time direction
            that is not committed.
            Affects only forward and frugal decoding schemes.
            Default value is ``d`` for forward scheme
            and ``1`` (``2*(d//2)``) for frugal scheme.
        :param merge_equivalent_boundary_nodes: whether to merge
            all nodes that represent the same boundary.
            This results in a decoding graph with
            as many boundary nodes as there are boundaries.
        :param parametrization: defines relative fault probabilities of
            1- and 2-qubit gates, and prep/measurement.
            Affects only circuit-level noise.
        :param demolition: whether measurement destroys the ancilla qubit state
            which hence needs to be initialized for next measurement cycle.
            Affects only circuit-level noise.
        :param monolingual: whether can prep/measure in only Z basis
            hence X-basis prep/measurement needs Hadamard gates.
            Affects only circuit-level noise.
        :param _merge_redundant_edges: whether to merge redundant boundary edges.
            Affects only circuit-level noise.
        """
        self._D = d
        self._MERGED_EQUIVALENT_BOUNDARY_NODES = merge_equivalent_boundary_nodes

        # INNER INIT START
        if 'batch' in scheme:
            for height in (commit_height, buffer_height):
                if height is not None:
                    raise ValueError(f"Cannot specify `{height=}` for `{scheme}` scheme.")
            if noise == 'code capacity':
                if window_height is not None:
                    raise ValueError(f"Cannot specify `window_height` for code capacity noise model.")
                h = 1
                self._N_EDGES, self._EDGES = self._code_capacity_edges(merge_equivalent_boundary_nodes)
                self._NOISE = CodeCapacity(self.EDGES)
            else:
                h = d if window_height is None else window_height
                if noise == 'phenomenological':
                    self._N_EDGES, self._EDGES = self._phenomenological_edges(
                        h,
                        False,
                        merge_equivalent_boundary_nodes=merge_equivalent_boundary_nodes,
                    )
                    self._NOISE = Phenomenological(self.EDGES)
                else:  # noise == 'circuit-level'
                    self._N_EDGES, self._EDGES, edge_dict, merges = self._circuit_level_edges(
                        h=h,
                        future_boundary=False,
                        _merge_redundant_edges=_merge_redundant_edges,
                        merge_equivalent_boundary_nodes=merge_equivalent_boundary_nodes,
                    )
                    self._NOISE = CircuitLevel(
                        fresh_edge_dict=edge_dict,
                        parametrization=parametrization,
                        demolition=demolition,
                        monolingual=monolingual,
                        fresh_merges=merges,
                    )
            self._SCHEME = Batch(self, h) if scheme == 'batch' else Global(self, h)
        else:
            if merge_equivalent_boundary_nodes:
                raise NotImplementedError("Yet to implement merging equivalent boundary nodes for the stream decoding schemes.")
            if window_height is not None:
                raise ValueError(f"Cannot specify `window_height` for the {scheme} decoding scheme.")
            if scheme == 'forward':
                if commit_height is None: commit_height = d
                if buffer_height is None: buffer_height = d
                scheme_class = Forward
            else:  # scheme == 'frugal'
                if commit_height is None: commit_height = 1
                if buffer_height is None: buffer_height = 2*(d//2)
                scheme_class = Frugal
            h = commit_height + buffer_height

            if noise == 'code capacity':
                raise TypeError(f"Code capacity incompatible with the {scheme} decoding scheme.")
            elif noise == 'phenomenological':
                self._N_EDGES, self._EDGES = self._phenomenological_edges(h, True)
                _, commit_edges = self._phenomenological_edges(commit_height, True)
                _,  fresh_edges = self._phenomenological_edges(commit_height, True, t_start=buffer_height)
                self._NOISE = Phenomenological(fresh_edges, all_edges=self._EDGES)
            else:  # noise == 'circuit-level'
                self._N_EDGES, self._EDGES, all_edge_dict, all_merges = self._circuit_level_edges(
                    h=h,
                    future_boundary=True,
                    _merge_redundant_edges=_merge_redundant_edges,
                )
                _, commit_edges, *_ = self._circuit_level_edges(
                    h=commit_height,
                    future_boundary=True,
                    _merge_redundant_edges=_merge_redundant_edges,
                )
                *_, fresh_edge_dict, fresh_merges = self._circuit_level_edges(
                    h=commit_height,
                    future_boundary=True,
                    _merge_redundant_edges=_merge_redundant_edges,
                    t_start=buffer_height,
                )
                self._NOISE = CircuitLevel(
                    fresh_edge_dict=fresh_edge_dict,
                    parametrization=parametrization,
                    demolition=demolition,
                    monolingual=monolingual,
                    fresh_merges=fresh_merges,
                    all_edge_dict=all_edge_dict,
                    all_merges=all_merges,
                )
            
            self._SCHEME = scheme_class(
                    self,
                    commit_height,
                    buffer_height,
                    commit_edges,
                )
        # INNER INIT END

        self._EDGES: tuple[Edge, ...]
        self._N_EDGES: int
        self._SCHEME: Scheme
        self._NOISE: Noise
        self._NODES: tuple[Node, ...] = tuple(set().union(*self.EDGES))
        
        ie: defaultdict[Node, set[Edge]] = defaultdict(set)
        for e in self.EDGES:
            for v in e:
                ie[v].add(e)
        self._INCIDENT_EDGES = dict(ie)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.D}, '{str(self.NOISE)}', scheme='{str(self.SCHEME)}')"
    
    @abc.abstractmethod
    def _code_capacity_edges(
        self,
        merge_equivalent_boundary_nodes: bool,
    ) -> tuple[int, tuple[Edge, ...]]:
        """Edges of G for code capacity noise model.
        
        
        :returns:
            * number of edges in G.
            * tuple of edges of G.
        """

    @abc.abstractmethod
    def _phenomenological_edges(
        self,
        h: int,
        future_boundary: bool,
        t_start=0,
        merge_equivalent_boundary_nodes=False,
    ) -> tuple[int, tuple[Edge, ...]]:
        """Edges of decoding graph G or viewing window W for phenomenological noise model.
        
        
        :param h: window height.
        :param future_boundary: whether top of graph has future boundary.
            True for W, False for G.
        :param t_start: the time index of the bottom layer.
        :param merge_equivalent_boundary_nodes: whether to merge
            all nodes that represent the same boundary.
        
        
        :return edge_count: Number of edges in the graph.
        :return edges: Tuple of edges of the graph.
        """

    @abc.abstractmethod
    def _circuit_level_edges(
            self,
            h: int,
            future_boundary: bool,
            _merge_redundant_edges: bool,
            t_start=0,
            merge_equivalent_boundary_nodes=False,
    ) -> tuple[
        int,
        tuple[Edge, ...],
        dict[EdgeType, tuple[Edge, ...]],
        dict[Edge, Edge] | None,
    ]:
        """Return inputs for ``noise.CircuitLevel``.
        
        :param h: window height.
        :param future_boundary: whether top of graph has future boundary.
            True for W, False for G.
        :param t_start: the time index of the bottom layer.
        :param merge_equivalent_boundary_nodes: whether to merge
            all nodes that represent the same boundary.
        :param _merge_redundant_edges: Whether to merge redundant boundary edges.
            Automatically ``True`` if ``merge_equivalent_boundary_nodes`` is ``True``.
        
        
        :return edge_count: Number of edges in the graph.
        :return edges: Tuple of edges of the graph which excludes the redundant edges if ``_merge_redundant_edges`` is ``True``.
        :return edge_dict: A map from edge type (i.e. orientation and location) to tuple of all edges of that type. Always includes redundant edges. Inter-key order matters as used by ``noise.forcers.ForceByEdge.force_error``.
        :return merges: A map each redundant edge to its substitute.
        """

    @abc.abstractmethod
    def _future_boundary_nodes(self, h: int) -> list[Node]:
        """Return list of boundary nodes at top of viewing window.
        
        :param h: window height.
        """

    @abc.abstractmethod
    def _redundant_boundary_nodes(self, h: int) -> list[Node]:
        """Return list of additional boundary nodes due to unmerged redundant edges.
        
        Used only by circuit-level noise.
        
        :param h: window height.
        """

    @property
    def D(self):
        """The code distance."""
        return self._D

    @property
    def MERGED_EQUIVALENT_BOUNDARY_NODES(self):
        """Whether all nodes that represent the same boundary are merged.
        This results in a decoding graph with
        as many boundary nodes as there are boundaries.
        """
        return self._MERGED_EQUIVALENT_BOUNDARY_NODES

    @property
    def SCHEME(self):
        """The decoding scheme."""
        return self._SCHEME

    @property
    def N_EDGES(self):
        """Number of edges in G."""
        return self._N_EDGES

    @property
    def EDGES(self):
        """A tuple of edges of G."""
        # Use tuple instead of generator so can repeatedly iterate through
        return self._EDGES

    @property
    def NODES(self):
        """A tuple of nodes of G."""
        return self._NODES

    @cached_property
    def NODE_COUNT(self):
        """Number of nodes in G."""
        return len(self.NODES)

    @property
    def NOISE(self):
        """The noise model.
        
        If the noise model is:
        
        * code capacity,
            then the decoding graph G represents the code, where each
            * detector represents a measure-Z qubit;
            * edge, a data qubit i.e. possible bitflip location.
        * phenomenological,
            then G represents w+1 measurement rounds of the code,
            where w is the decoding window height and each
            * detector represents the difference between two consecutive measurements
                of a measure-Z qubit at a given point in time;
            * horizontal edge, a possible time at which a data qubit could bitflip;
            * vertical edge, a possible faulty measurement
                (i.e. measure-Z qubit recording the wrong parity with some probability q)
                location.
        * circuit-level,
            then G represents the same thing as for phenomenological noise,
            but each edge represents a possible pair of defects
            that could have resulted from one fault.
        """
        return self._NOISE

    @property
    def TIME_AXIS(self):
        """The axis that represents time."""
        return self._TIME_AXIS

    @property
    def LONG_AXIS(self):
        """The axis whose index runs from -1 to d-1 inclusive."""
        return self._LONG_AXIS

    @cached_property
    def DIMENSION(self):
        """The number of dimensions that G is embedded in."""
        return self._CODE_DIMENSION + (str(self.NOISE) != 'code capacity')

    @property
    def INCIDENT_EDGES(self):
        """A map from each node to a set of incident edges."""
        return self._INCIDENT_EDGES
    # Tried construction via manually changing indices by 1
    # so need not iterate through `self.EDGES`,
    # but this is ~twice as slow.
    
    @cached_property
    def DETECTORS(self):
        """A tuple of all detectors of G."""
        return tuple(v for v in self.NODES if not self.is_boundary(v))

    @cached_property
    def DETECTOR_COUNT(self):
        """Number of detectors in G."""
        return len(self.DETECTORS)

    @cached_property
    def BOUNDARY_NODES(self):
        """A tuple of all boundary nodes of G."""
        return tuple(v for v in self.NODES if self.is_boundary(v))

    def is_boundary(self, v: Node):
        """Determine whether ``v`` a boundary node."""
        return self.SCHEME.is_boundary(v)

    def neighbors(self, v: Node):
        """Return neighbors of ``v``.
        
        Never actually used but if do,
        change to @cached_property.
        """
        ans: set[Node] = set()
        for e in self.INCIDENT_EDGES[v]:
            ans.update(e)
        ans.remove(v)
        return ans

    @staticmethod
    def traverse_edge(e: Edge, u: Node):
        """Return node at other end of edge ``e``."""
        v = e[1] if u == e[0] else e[0]
        return v
    
    def raise_node(self, v: Node, delta_t: int = 1) -> Node:
        """Move ``v`` up by ``delta_t``."""
        # tuple unpacking is marginally slower
        new_v = list(v)
        new_v[self.TIME_AXIS] += delta_t
        return tuple(new_v)
    
    def raise_edge(self, e: Edge, delta_t: int = 1) -> Edge:
        """Move ``e`` up by ``delta_t``."""
        return tuple(self.raise_node(v, delta_t) for v in e) # type: ignore

    def make_error(self, noise_level: float, exclude_future_boundary: bool = False):
        """Sample edges from freshly discovered region.
        
        
        :param noise_level: characteristic noise level if circuit-level noise;
            else, bitflip probability.
            Should be in [0, 1], though no check is done to ensure this.
        :param exclude_future_boundary: whether to exclude future boundary edges
            from being sampled for the new error in the freshly discovered region.
            Set to ``True`` if you want to emulate the end of a memory experiment,
            where the data qubits are measured and the last syndrome sheet is obtained
            from classically taking parities of these measurement outcomes
            (so there is no measurement error).
        
        
        :return error: The set of bitflipped edges in the freshly discovered region.
            Each edge bitflips with probability defined by its multiplicity
            if circuit-level noise; else, with probability ``noise_level``.
        """
        error = self.NOISE.make_error(noise_level)
        if exclude_future_boundary:
            excluded_edges: set[Edge] = set()
            for e in error:
                for v in e:
                    if v[self.TIME_AXIS] == self.SCHEME.WINDOW_HEIGHT:
                        excluded_edges.add(e)
            error.difference_update(excluded_edges)
        return error

    def get_syndrome(self, error: set[Edge]):
        """Get syndrome from error configuration.
        
        
        :param error: a set of bitflipped edges.
        
        
        :returns syndrome: a set of defects.
        """
        return self.get_verbose_syndrome(error).difference(self.BOUNDARY_NODES)

    def get_verbose_syndrome(self, error: set[Edge]):
        """Get syndrome, treating boundary nodes as detectors too.
        
        
        :param error: a set of bitflipped edges.
        
        
        :returns verbose_syndrome: a set of defects.
        """
        # Note: Implementing `verbose_syndrome` as a set we add to and remove from is
        # empirically faster than as a dictionary w/ a key for each measure-Z qubit
        # and Booleans as values, which we flip back and forth (for d=29, noise_level=0.11:
        # 3.03(4) ms < 3.26(5) ms).
        verbose_syndrome: set[Node] = set()
        for e in error:
            verbose_syndrome.symmetric_difference_update(e)
        return verbose_syndrome

    @staticmethod
    def compose_errors(*errors: set[Edge]):
        """Sequentially compose any number of errors.
        
        :param errors: a tuple (error1, error2, ...) where each
            error is a set of bitflipped edges.
        
        
        :returns composition: A set of edges representing
            the sequential composition of all errors in ``errors``.
        """
        composition: set[Edge] = set()
        for error in errors:
            composition ^= error
        return composition

    def get_logical_error(self, leftover: set[Edge]):
        """Whether leftover implements logical X.
        
        
        :param leftover: a set of bitflipped edges.
        
        
        :returns logical_error: logical error count parity
            if scheme is batch else logical error count
            if scheme is scheme is global batch else logical error count in commit region
            if scheme is forward.
        """
        return self.SCHEME.get_logical_error(leftover)
    
    @abc.abstractmethod
    def index_to_id(self, index: Node) -> int:
        """Return unique ID of node."""

    @cached_property
    def GRAPH(self):
        """A NetworkX graph of G.
        
        Notes: caching so only need make it once provides noticeable speedup!
        """
        import networkx as nx
        g = nx.Graph(
            self.EDGES,
            D=self.D,
            SCHEME=str(self.SCHEME),
        )
        for v in g.nodes:
            g.nodes[v]['is_boundary'] = self.is_boundary(v)
        return g

    def draw(
        self,
        error: set[Edge] | None = None,
        syndrome: set[Node] | None = None,
        x_offset=constants.DEFAULT_X_OFFSET,
        with_labels: bool | None = None,
        nodelist: Sequence[Node] | None = None,
        node_size: float | Collection[float] | None = None,
        width: float | Collection[float] | None = None,
        boundary_color=constants.BLUE,
        defect_color=constants.RED,
        nondefect_color=constants.GREEN,
        error_color=constants.RED,
        non_error_color='k',
        **kwargs_for_networkx_draw,
    ):
        """Draw G using matplotlib.
        
        
        :param error: a set of edges. Default is empty set.
        :param syndrome: a set of defects. Default is that produced by ``error``.
        :param x_offset: the ratio of out-of-screen to along-screen distance.
        :param with_labels: whether to draw labels on each node.
        :param nodelist: the subset of nodes to draw. Default is all nodes i.e. ``self.NODES``.
        :param node_size: size of nodes.
        :param width: line width of edges.
        :param boundary_color: string specifying color of boundary nodes.
        :param defect_color: string specifying color of defects.
        :param nondefect_color: string specifying color of nondefects.
        :param error_color: string specifying color of bitflipped edges.
        :param non_error_color: string specifying color of unflipped edges.
        :param kwargs_for_networkx_draw: modifies/adds keyword arguments to ``networkx.draw()``.
        
        Draws: G, where
        * bitflipped edges thick red; else, thin black
        * boundary nodes blue; defects red; else, green.
        
        Output: The NetworkX graph of G.
        """
        import networkx as nx
        # get error and syndrome
        if error is None: error = set()
        if syndrome is None: syndrome = self.get_syndrome(error)

        # get graph and kwargs for nx.draw()
        g = self.GRAPH
        pos = self.get_pos(x_offset)
        # if d small enough, can draw labels on each node
        if with_labels is None:
            with_labels = (self.D <= 3) if self.DIMENSION == 3 else (self.D <= 7)
        if nodelist is None:
            nodelist = self.NODES
        if node_size is None:
            node_size = constants.DEFAULT if with_labels else constants.SMALL
        node_color = self.get_node_color(
            syndrome,
            boundary_color=boundary_color,
            defect_color=defect_color,
            nondefect_color=nondefect_color,
            nodelist=nodelist,
        )
        if width is None:
            width = [
                constants.WIDE if e in error else
                constants.MEDIUM_THIN if with_labels else
                constants.THIN if self.DIMENSION == 3 else
                0 for e in self.EDGES
            ]
        edge_color = [
            error_color if e in error else
            non_error_color for e in self.EDGES
        ]
        nx.draw(
            g,
            pos=pos,
            with_labels=with_labels,
            nodelist=nodelist,
            edgelist=self.EDGES,
            node_size=node_size,
            node_color=node_color,
            width=width,
            edge_color=edge_color,
            **kwargs_for_networkx_draw,
        )
        return g

    @abc.abstractmethod
    def get_pos(
        self,
        x_offset: float = constants.DEFAULT_X_OFFSET,
        nodelist: None | Iterable[Node] = None,
    ) -> dict[Node, Coord]:
        """Compute coordinates of each node G for ``draw()``.
        
        
        :param x_offset: the ratio of out-of-screen to along-screen distance.
        :param nodelist: the nodes to draw. Default is ``self.NODES``.
        
        :returns pos: a dictionary where each key a node index; value, position coordinate.
            E.g. for the surface code with perfect measurements,
            convert each matrix index to position coords via
            (i, j) -> (x, y) = (j, -i).
        """

    def get_node_color(
            self,
            syndrome: Container[Node],
            boundary_color=constants.BLUE,
            defect_color=constants.RED,
            nondefect_color=constants.GREEN,
            nodelist: Iterable[Node] | None = None,
            show_boundary_defects=True,
    ):
        """Return a list of colors each node should be for ``draw()``.
        
        
        :param syndrome: a set of defects.
        :param boundary_color: string specifying color of boundary nodes.
        :param defect_color: string specifying color of defects.
        :param nondefect_color: string specifying color of nondefects.
        :param nodelist: the nodes to draw. Default is all nodes i.e. ``self.NODES``.
        :param show_boundary_defects: whether to boundary nodes can be defects.
        
        :returns node_color: A list of colors for each node in ``nodelist``.
        """
        if nodelist is None:
            nodelist = self.NODES
        if show_boundary_defects:
            node_color = [
                defect_color if v in syndrome else
                boundary_color if self.is_boundary(v) else
                nondefect_color for v in nodelist
            ]
        else:
            node_color = [
                boundary_color if self.is_boundary(v) else
                defect_color if v in syndrome else
                nondefect_color for v in nodelist
            ]
        return node_color
    

class Decoder(abc.ABC):
    """Base class for decoders."""

    @property
    def CODE(self):
        """The code to be decoded."""
        return self._CODE

    def __init__(self, code: Code):
        """Input: ``code`` the code to be decoded."""
        self._CODE = code
        self.correction: set[Edge]
        """A set of edges comprising the correction."""

    def reset(self):
        """Factory reset."""
        self.correction.clear()

    @abc.abstractmethod
    def decode(
            self,
            syndrome: set[Node],
            **kwargs,
    ) -> None | int | tuple[int, int]:
        """Decode syndrome.
        
        :param syndrome: the set of defects.
        :param draw: whether to draw.
        :param log_history: whether keep track of state history.
        
        Correction stored in ``self.correction``.
        """

    @abc.abstractmethod
    def draw_decode(self, **kwargs_for_networkx_draw):
        """Draw all stages of decoding.
        
        Input: ``kwargs_for_networkx_draw`` passed to ``NetworkX.draw``
        e.g. ``margins=(0.1, 0.1)``.
        """

    def subset_sample(self, noise_level: float, n: int, tol: float = 5e-1):
        """Simulate decoding cycles for each error subset.
        
        
        :param noise_level: noise level.
        :param n: number of decoding cycles per subset.
        :param tol: how much cutoff error we can tolerate,
            as a fraction of the mean.
        
        
        :returns: A pandas DataFrame indexed by ``weight`` with columns ``['subset prob', 'survival prob', 'm', 'n']``.
        """
        subset_probs = self.CODE.NOISE.subset_probabilities(noise_level)
        mean, mn, weights = 0, [], []
        for weight in subset_probs.index:
            weights.append(weight)
            weightless = weight==0 if isinstance(weight, int) else all(w==0 for w in weight)
            if weightless:
                mn.append((np.nan, np.nan))
            else:
                m, shots = self.CODE.SCHEME.sim_cycles_given_weight(self, weight, n)
                mn.append((m, shots))
                mean += subset_probs.loc[weight, 'subset prob'] * m / shots # type: ignore
                if mean * tol > subset_probs.loc[weight, 'survival prob']: # type: ignore
                    break
        if type(subset_probs.index) is pd.MultiIndex:
            index = pd.MultiIndex.from_tuples(weights)
        else:
            index = pd.Index(weights)
        subset_failures = pd.DataFrame(mn, index=index, columns=['m', 'n'])
        return pd.concat([subset_probs, subset_failures], axis=1, join='inner')


"""Module for Snowflake decoder."""

# from localuf.decoders.snowflake.constants import Stage
from localuf.decoders.snowflake.constants import RESET
from localuf.decoders.snowflake.main import \
    Snowflake, _Node, _Edge, NodeFriendship, TopSheetFriendship, \
    NothingFriendship, Friendship, _SlowMerger, \
    _FastMerger, _Merger, EdgeContact, \
    FloorContact, _Contact


from localuf.decoders import Snowflake
from localuf.decoders.snowflake import _Node
from localuf.type_aliases import Node
from localuf.decoders.snowflake.constants import RESET
from localuf.decoders._base_uf import direction


def test_start(sfn3: _Node):
    sfn3.pointer = 'D'
    sfn3.UNROOTER.start()
    assert sfn3.cid == RESET
    assert sfn3.pointer == 'C'


def test_flooding(syncing_flooding_objects: tuple[
    Snowflake, tuple[Node, Node, Node], tuple[_Node, _Node, _Node]
]):
    snowflake3, _, (west, center, east) = syncing_flooding_objects
    access: dict[direction, _Node] = {
        'W': west,
        'E': east,
    }

    # TEST FINISH UNROOTING
    center.cid = RESET
    center.next_cid = west.ID
    center.UNROOTER.flooding_whole()
    assert center.busy
    assert center.next_cid == center.ID
    assert center.next_unrooted
    center.reset()

    # TEST START UNROOTING
    west.cid = RESET
    center.access = access
    center.pointer = 'W'
    center.UNROOTER.flooding_whole()
    assert center.busy
    assert center.next_cid == RESET
    assert center.pointer == 'C'
    snowflake3.reset()

    # TEST UPDATE `pointer`, `cid`
    center.access = access
    center.UNROOTER.flooding_half()
    assert center.busy
    assert center.pointer == 'W'
    assert center.next_cid == west.ID


import pytest

from localuf.codes import Surface
from localuf.decoders._base_uf import BaseUF
from localuf.constants import Growth
from localuf.type_aliases import Node


class ConcreteBaseUF(BaseUF):

    correction = set()
    
    @property
    def _FIG_HEIGHT(self): return 0
    
    @property
    def _FIG_WIDTH(self): return self._FIG_HEIGHT

    def decode(self):
        """Override abstract method."""

    def draw_decode(self):
        """Override abstract method."""


@pytest.fixture
def uf5F(sf5F: Surface):
    return ConcreteBaseUF(sf5F)


@pytest.fixture
def uf7F(sf7F: Surface):
    return ConcreteBaseUF(sf7F)


@pytest.fixture
def uf3T(sf3T: Surface):
    return ConcreteBaseUF(sf3T)


def test_reset(uf7F: BaseUF):
    for e in uf7F.growth.keys():
        uf7F.growth[e] = Growth.FULL
    uf7F.erasure
    uf7F.reset()
    assert all(growth == 0 for growth in uf7F.growth.values())
    with pytest.raises(AttributeError, match="erasure"):
        del uf7F.erasure


class TestErasureProperty:

    def test_empty(self, uf7F: BaseUF):
        assert uf7F.erasure == set()

    def test_with_full_growth(self, uf7F: BaseUF):
        uf7F.growth[(0, -1), (0, 0)] = Growth.FULL
        assert uf7F.erasure == {((0, -1), (0, 0))}

    def test_with_half_growth(self, uf7F: BaseUF):
        uf7F.growth[(0, -1), (0, 0)] = Growth.HALF
        assert uf7F.erasure == set()

    def test_with_mixed_growth(self, uf7F: BaseUF, uvw: tuple[Node, Node, Node]):
        u, v, w = uvw
        assert uf7F.erasure == set()
        uf7F.growth[u, v] = Growth.HALF
        uf7F.growth[v, w] = Growth.FULL
        # erasure a cached_property so need delete attribute to run method again
        del uf7F.erasure
        assert uf7F.erasure == {(v, w)}


class TestSwimDistance:

    def test_no_clusters_5F(self, uf5F: BaseUF):
        assert uf5F.swim_distance() == uf5F.CODE.D

    @pytest.mark.parametrize("i", range(5))
    @pytest.mark.parametrize("max_j", range(5))
    def test_percolated_5F(self, uf5F: BaseUF, i: int, max_j: int):
        for j in range(-1, max_j-1):
            uf5F.growth[(i, j), (i, j+1)] = Growth.FULL
        assert uf5F.swim_distance() == 5-max_j

    @pytest.mark.parametrize("i", range(5))
    @pytest.mark.parametrize("max_j", range(5))
    def test_half_percolated_5F(self, uf5F: BaseUF, i: int, max_j: int):
        for j in range(-1, max_j-1):
            uf5F.growth[(i, j), (i, j+1)] = Growth.HALF
        assert uf5F.swim_distance() == 5 - max_j/2

    def test_no_clusters_3T(self, uf3T: BaseUF):
        assert uf3T.swim_distance() == uf3T.CODE.D

    @pytest.mark.parametrize("i", range(3))
    @pytest.mark.parametrize("max_j", range(3))
    @pytest.mark.parametrize("t", range(3))
    def test_percolated_3T(self, uf3T: BaseUF, i: int, max_j: int, t: int):
        for j in range(-1, max_j-1):
            uf3T.growth[(i, j, t), (i, j+1, t)] = Growth.FULL
        assert uf3T.swim_distance() == 3-max_j

    @pytest.mark.parametrize("i", range(3))
    @pytest.mark.parametrize("max_j", range(3))
    @pytest.mark.parametrize("t", range(3))
    def test_half_percolated_3T(self, uf3T: BaseUF, i: int, max_j: int, t: int):
        for j in range(-1, max_j-1):
            uf3T.growth[(i, j, t), (i, j+1, t)] = Growth.HALF
        assert uf3T.swim_distance() == 3 - max_j/2


"""
localuf (Local Union--Find)
============================

A Python package to simulate and visualise decoders for CSS codes.

Available code classes:

* Surface
* Repetition

Available modules:

* decoders
* plot
* sim

Available functions:

* get_failure_data
* get_log_runtime_data
* get_stats
* get_failure_data_from_subset_sample
* add_ignored_timesteps
"""


# print(f'Invoking __init__.py for {__name__}')
# below correct (see https://github.com/psf/requests/blob/v2.23.0/requests/__init__.py#L112)
from localuf.codes import Repetition, Surface
from localuf import decoders
from localuf import plot
from localuf import sim
from localuf.data_processors import get_failure_data, get_log_runtime_data, \
    get_stats, get_failure_data_from_subset_sample, add_ignored_timesteps


import itertools
import re

import pytest

from localuf.noise import CircuitLevel
from localuf.noise.forcers import ForceByEdge
from localuf.type_aliases import Edge


def test_force_error(toy_cl: CircuitLevel, e_westmost: tuple[Edge, Edge]):
    fbe = ForceByEdge(toy_cl._FRESH_EDGES)
    for weight in itertools.product((0, 1), repeat=2):
        error = fbe.force_error(weight)
        assert error == {e for e, w in zip(e_westmost, weight) if w}
    with pytest.raises(
        ValueError,
        match='Sample larger than population or is negative',
    ):
        fbe.force_error((2, 0))
    with pytest.raises(
        ValueError,
        match=re.escape("len(weight)=3 != 2=number of edge subsets")
    ):
        fbe.force_error((0, 0, 0))


"""Provide ``get_heights`` function."""

from collections.abc import Callable

from localuf.type_aliases import DecodingScheme


def get_heights(
        d: int,
        global_batch_slenderness: int,
        scheme: DecodingScheme,
        get_commit_height: Callable[[int], int] | None,
        get_buffer_height: Callable[[int], int] | None,
):
    """Get height inputs for ``Code.__init__`` based on decoding scheme and multipliers.
    
    
    :param d: code distance.
    :param global_batch_slenderness: the slenderness in case ``scheme == 'global batch'``.
    :param scheme: decoding scheme.
    :param get_commit_height: a function with input ``d`` that outputs commit height.
    :param get_buffer_height: a function with input ``d`` that outputs buffer height.
    """
    window_height = d*global_batch_slenderness if scheme == 'global batch' else None

    if get_commit_height is None:
        if scheme == 'forward':
            commit_height = d
        elif scheme == 'frugal':
            commit_height = 1
        else:  # 'batch' in scheme
            commit_height = None
    else:
        commit_height = get_commit_height(d)

    if get_buffer_height is None:
        if scheme == 'forward':
            buffer_height = d
        elif scheme == 'frugal':
            buffer_height = 2*(d//2)
        else:  # 'batch' in scheme
            buffer_height = None
    else:
        buffer_height = get_buffer_height(d)

    return window_height, commit_height, buffer_height


from localuf import Surface
from localuf.decoders.luf import Macar

def test_merging_never_relay_from_boundary(sf3F: Surface):
    decoder = Macar(sf3F)
    syndrome = {
        (0, 0),
        (2, 0),
    }
    n_steps = decoder.validate(syndrome)
    assert n_steps == 12
    capital_I = {
        ((0, -1), (0, 0)),
        ((0, 0), (0, 1)),
        ((0, 0), (1, 0)),
        ((1, 0), (2, 0)),
        ((2, -1), (2, 0)),
        ((2, 0), (2, 1)),
    }
    assert decoder.erasure == capital_I


from collections.abc import Callable
import itertools

import pytest

from localuf.type_aliases import Node
from localuf.constants import Growth
from localuf.decoders.node_uf import NodeUF, _NodeCluster

@pytest.fixture
def uf7F(sf7F):
    uf_graph = NodeUF(sf7F)
    uf_graph.history = []
    return uf_graph

@pytest.fixture
def dynamic_uf7F(sf7F):
    uf_graph = NodeUF(sf7F, dynamic=True)
    uf_graph.history = []
    return uf_graph

@pytest.fixture
def first_changed_edges():
    return {
        ((0, -1), (0, 0)),
        ((0, 0), (0, 1)),
        ((0, 0), (1, 0)),
        ((0, 1), (0, 2)),
        ((0, 1), (1, 1)),
        ((0, 3), (0, 4)),
        ((0, 4), (0, 5)),
        ((0, 4), (1, 4)),
        ((1, 0), (2, 0)),
        ((1, 3), (1, 4)),
        ((1, 4), (1, 5)),
        ((1, 4), (2, 4)),
        ((2, -1), (2, 0)),
        ((2, 0), (2, 1)),
        ((2, 0), (3, 0)),
        ((2, 4), (3, 4)),
        ((3, 3), (3, 4)),
        ((3, 3), (4, 3)),
        ((3, 4), (3, 5)),
        ((3, 4), (4, 4)),
        ((4, 2), (4, 3)),
        ((4, 3), (4, 4)),
        ((4, 3), (5, 3)),
        ((4, 4), (4, 5)),
        ((4, 4), (5, 4))
    }

@pytest.fixture
def uf_after_union(
    uf7F: NodeUF,
    get_uf_after_union: Callable[[NodeUF], tuple[NodeUF, _NodeCluster, _NodeCluster]],
    syndrome7F: set[Node],
):
    uf7F.load(syndrome7F)
    return get_uf_after_union(uf7F)

def test_validate_static(
        uf7F: NodeUF,
        validated_static_erasure,
        syndrome7F,
):
    uf7F.validate(syndrome7F)
    assert len(uf7F.clusters) == 26
    assert uf7F.active_clusters == set()
    assert uf7F.erasure == validated_static_erasure

def test_validate_dynamic(
        dynamic_uf7F: NodeUF,
        validated_dynamic_erasure,
        syndrome7F,
):
    dynamic_uf7F.validate(syndrome7F)
    assert len(dynamic_uf7F.clusters) == 27
    assert dynamic_uf7F.active_clusters == set()
    assert dynamic_uf7F.erasure == validated_dynamic_erasure

def test_growth_round_static(
        uf7F: NodeUF,
        syndrome7F,
        first_changed_edges,
):
    uf7F.load(syndrome7F)
    log_history = True
    uf7F._growth_round(log_history)
    assert uf7F.history[0].changed_edges == first_changed_edges
    while uf7F.active_clusters:
        assert all(v in {
            Growth.UNGROWN,
            Growth.HALF,
            Growth.FULL
        } for v in uf7F.growth.values())
        uf7F._growth_round(log_history)

def test_growth_round_dynamic(
        dynamic_uf7F: NodeUF,
        syndrome7F,
        first_changed_edges,
):
    dynamic_uf7F.load(syndrome7F)
    log_history = True
    dynamic_uf7F._growth_round(log_history)
    assert dynamic_uf7F.history[0].changed_edges == first_changed_edges
    while dynamic_uf7F.active_clusters:
        assert all(v in {
            Growth.BURNT,
            Growth.UNGROWN,
            Growth.HALF,
            Growth.FULL
        } for v in dynamic_uf7F.growth.values())
        dynamic_uf7F._growth_round(log_history)

def test_grow(uf7F: NodeUF, fixture_test_grow):
    sq, c = fixture_test_grow(uf7F)

    # test ex-frontier nodes removed
    for _ in itertools.repeat(None, 2):
        uf7F._growth_round(
            log_history=False,
            clusters_to_grow={c}
        )
        c = uf7F.clusters[uf7F._find(sq)]
    assert c.frontier == {
        (1, 0),
        (2, 1),
        (3, 0),
    }

def test_union(uf_after_union, fixture_test_union):
    fixture_test_union(uf_after_union)
    _, larger, smaller = uf_after_union
    assert larger.frontier == {smaller.root}


import pytest

from localuf import Surface
from localuf.type_aliases import Node
from localuf.decoders.luf import Macar, Actis

@pytest.fixture(name="luf", params=[Macar, Actis], ids=["Macar", "Actis"])
def _luf(sf: Surface, request):
    return request.param(sf)

@pytest.fixture
def macar(sf: Surface):
    return Macar(sf)

@pytest.fixture
def actis(sf: Surface):
    return Actis(sf)

@pytest.fixture(name="luf3F", params=[Macar, Actis], ids=["Macar", "Actis"])
def _luf3F(sf3F: Surface, request):
    return request.param(sf3F)

@pytest.fixture(name="luf3T", params=[Macar, Actis], ids=["Macar", "Actis"])
def _luf3T(sf3T: Surface, request):
    return request.param(sf3T)

@pytest.fixture
def macar3F(sf3F: Surface):
    return Macar(sf3F)

@pytest.fixture
def actis3Fu(sf3F: Surface):
    return Actis(sf3F, _optimal=False)

@pytest.fixture
def macar3T(sf3T: Surface):
    return Macar(sf3T)

@pytest.fixture
def actis3Tu(sf3T: Surface):
    return Actis(sf3T, _optimal=False)

@pytest.fixture
def macar5F(sf5F: Surface):
    return Macar(sf5F)

@pytest.fixture
def actis5Fu(sf5F: Surface):
    return Actis(sf5F, _optimal=False)

@pytest.fixture
def actis5T(sf5T: Surface):
    return Actis(sf5T)

@pytest.fixture
def macar5T(sf5T: Surface):
    return Macar(sf5T)

@pytest.fixture(name="ns3F", params=[Macar, Actis], ids=["Macar", "Actis"])
def _ns3F(sf3F, request):
    decoder = request.param(sf3F)
    return decoder.NODES

@pytest.fixture(name="ns3T", params=[Macar, Actis], ids=["Macar", "Actis"])
def _ns3T(sf3T, request):
    decoder = request.param(sf3T)
    return decoder.NODES

@pytest.fixture
def actis_nodes(actis: Actis):
    return actis.NODES

@pytest.fixture
def actis5Fu_nodes(actis5Fu: Actis):
    return actis5Fu.NODES

@pytest.fixture
def actis5T_nodes(actis5T: Actis):
    return actis5T.NODES

@pytest.fixture
def syndrome5F() -> set[Node]:
    return {
        (0, 0),
        (0, 1),
        (0, 2),
        (4, 1),
    }

@pytest.fixture
def syndrome5T() -> set[Node]:
    return {
        (0, 0, 0),
        (0, 1, 1),
        (0, 2, 2),
        (4, 1, 3),
    }


import pytest
from localuf.noise import CircuitLevel

from localuf.type_aliases import Edge, EdgeType


@pytest.fixture
def e_westmost() -> tuple[Edge, Edge]:
    return tuple(((1, -1, t), (1, 0, t)) for t in (0, 1)) # type: ignore


@pytest.fixture
def toy_edge_dict(e_westmost: tuple[Edge, Edge]):
    f = ((1, -1, 0), (1, 0, 1))
    g = ((0, -1, 0), (1, 0, 1))
    return {
        'E westmost': e_westmost,
        'EU west corners': (f,),
        'SEU': (g,),
    }


@pytest.fixture
def toy_merges(toy_edge_dict: dict[EdgeType, tuple[Edge, ...]]):
    _, e1 = toy_edge_dict['E westmost']
    f, = toy_edge_dict['EU west corners']
    g, = toy_edge_dict['SEU']
    return {f: e1, g: e1}


@pytest.fixture
def toy_cl(
    toy_edge_dict: dict[EdgeType, tuple[Edge, ...]],
    toy_merges: dict[Edge, Edge],
):
    return CircuitLevel(
        fresh_edge_dict=toy_edge_dict,
        parametrization='balanced',
        demolition=False,
        monolingual=False,
        fresh_merges=toy_merges,
    )


import pytest


from localuf import Surface, Repetition
from localuf._schemes import _Streaming
from localuf.noise.main import Phenomenological


class TestEdges:


    def test_batch(self, sf3T: Surface):
        model: Phenomenological = sf3T.NOISE  # type: ignore
        assert model.FRESH_EDGES == sf3T.EDGES


    @pytest.mark.parametrize("rp", ["rp3_forward", "rp3_frugal"])
    def test_nonbatch(self, rp: Repetition, request):
        rp = request.getfixturevalue(rp)
        scheme: _Streaming = rp.SCHEME # type: ignore
        buffer_height = scheme.WINDOW_HEIGHT - scheme._COMMIT_HEIGHT
        _, fresh_edges = rp._phenomenological_edges(
            scheme._COMMIT_HEIGHT,
            True,
            t_start=buffer_height,
        )
        model: Phenomenological = rp.NOISE  # type: ignore
        assert model.FRESH_EDGES == fresh_edges


from pymatching import Matching
import pytest

from localuf._base_classes import Code
from localuf.codes import Surface
from localuf.decoders import MWPM
from localuf.type_aliases import Node


@pytest.mark.parametrize("noise_level", [None, 1e-1, 1e-2])
def test_get_matching(sfCL: Surface, noise_level: float):
    decoder = MWPM(sfCL)
    matching = decoder.get_matching(noise_level=noise_level)
    assert matching.num_edges == sfCL.N_EDGES
    assert matching.num_detectors == sfCL.D * (sfCL.D-1) * sfCL.SCHEME.WINDOW_HEIGHT
    assert matching.num_fault_ids == 1
    assert_all_edge_data_correct(sfCL, matching, noise_level, decoder._DETECTOR_TO_INT)


def assert_all_edge_data_correct(
        code: Code,
        matching: Matching,
        noise_level: float,
        detector_to_int: dict[Node, int],
):
    """Assert all edge data are correct according to `noise_level`."""
    edge_weights = code.NOISE.get_edge_weights(noise_level)
    for (u, v), (flip_probability, weight) in edge_weights.items():
        if u in code.BOUNDARY_NODES:
            edge_data = matching.get_boundary_edge_data(detector_to_int[v])
        elif v in code.BOUNDARY_NODES:
            edge_data = matching.get_boundary_edge_data(detector_to_int[u])
        else:
            edge_data = matching.get_edge_data(
                node1=detector_to_int[u],
                node2=detector_to_int[v],
            )
        assert edge_data['fault_ids'] == ({0} if (u[code.LONG_AXIS] == -1) else set())
        assert edge_data['weight'] == weight
        assert edge_data['error_probability'] == flip_probability


import itertools
from unittest import mock

import pytest

from localuf.decoders.luf import LUF, Controller, Macar

def test_CONTROLLER_attribute(luf: LUF):
    assert type(luf.CONTROLLER) is Controller

def test_nodes_attribute(luf: LUF):
    d = luf.CODE.D
    assert type(luf.NODES.dc) is dict
    if luf.CODE.DIMENSION == 2:
        assert len(luf.NODES.dc) == d * (d+1)
    else:
        assert luf.CODE.DIMENSION == 3
        assert len(luf.NODES.dc) == d**2 * (d+1)

@pytest.mark.parametrize("prop", [
    "CONTROLLER",
    "NODES",
    "VISIBLE",
    "_pointer_digraph",
])
def test_property_attributes(test_property, luf: LUF, prop):
    test_property(luf, prop)

@pytest.mark.parametrize("decoder, syndrome", [
    ("macar5F", "syndrome5F"),
    ("macar5T", "syndrome5T"),
])
def test_validate_draw_True(decoder, syndrome, request):
    """From https://miguendes.me/how-to-use-fixtures-as-arguments-in-pytestmarkparametrize."""
    decoder = request.getfixturevalue(decoder)
    syndrome = request.getfixturevalue(syndrome)
    n_steps = decoder.validate(syndrome, draw=True)
    assert n_steps == len(decoder.history)

@pytest.mark.parametrize("sf, syndrome", [
    ("sf5F", "syndrome5F"),
    ("sf5T", "syndrome5T"),
])
def test_validate_draw_consistency(sf, syndrome, request):
    sf = request.getfixturevalue(sf)
    syndrome = request.getfixturevalue(syndrome)
    decoderF = Macar(sf)
    decoderT = Macar(sf)
    nF = decoderF.validate(syndrome)
    nT = decoderT.validate(syndrome, draw=True)
    assert nF == nT

def test_advance_physicals(macar: Macar):
    n_nodes = len(macar.CODE.NODES)
    with (
        mock.patch("localuf.decoders.luf.MacarNode.advance") as vn_advance,
        mock.patch("localuf.decoders.luf.Controller.advance") as c_advance,
    ):
        macar._advance()
        assert vn_advance.call_count == n_nodes
        c_advance.assert_called_once_with()

def test_advance_unphysicals_growth(macar: Macar):
    n_nodes = len(macar.CODE.NODES)
    with mock.patch("localuf.decoders.luf.MacarNode.update_access") as ua:
        macar._advance()
        assert ua.call_count == n_nodes

def test_advance_unphysicals_merge(macar: Macar):
    n_nodes = len(macar.CODE.NODES)
    macar._advance()
    with mock.patch("localuf.decoders.luf.MacarNode.update_after_merge_step") as step:
        macar._advance()
        assert step.call_count == n_nodes

def test_advance_unphysicals_sync(macar: Macar):
    n_nodes = len(macar.CODE.NODES)
    for _ in itertools.repeat(None, 3):
        macar._advance()
    with mock.patch("localuf.decoders.luf.MacarNode.update_after_sync_step") as uass:
        macar._advance()
        assert uass.call_count == n_nodes

@pytest.mark.parametrize("sf, syndrome", [
    ("sf5F", "syndrome5F"),
    ("sf5T", "syndrome5T"),
])
def test_reset(sf, syndrome, request):
    sf = request.getfixturevalue(sf)
    syndrome = request.getfixturevalue(syndrome)
    decoder_1 = Macar(sf)
    decoder_2 = Macar(sf)
    decoder_2.validate(syndrome, draw=True)
    decoder_2.syndrome
    decoder_2.reset()
    with pytest.raises(AttributeError, match="syndrome"):
        del decoder_2.syndrome
    assert dir(decoder_1) == dir(decoder_2)
    assert decoder_1.growth == decoder_2.growth
    assert decoder_1.syndrome == decoder_2.syndrome
    assert decoder_1.CONTROLLER.stage == decoder_2.CONTROLLER.stage
    assert not hasattr(decoder_2, 'history')


from localuf.decoders.luf import Controller
from localuf.decoders.luf.constants import Stage


def test_stage_symbol(c3: Controller):
    assert str(c3.stage) == 'G'
    c3.stage = Stage.MERGING
    assert str(c3.stage) == 'M'


"""Module for functions to process numerical data, mainly from ``sim``."""

from collections.abc import Callable
from typing import Type

import numpy as np
import numpy.typing as npt
import pandas as pd
from pandas import DataFrame
import statsmodels.api as sm
from statsmodels.stats.proportion import proportion_confint

from localuf._base_classes import Code
from localuf.type_aliases import NoiseModel
from localuf.constants import STANDARD_ERROR_ALPHA


def get_failure_data(
        data: DataFrame,
        p_slice=slice(None),
        alpha=STANDARD_ERROR_ALPHA,
        method='wilson',
) -> DataFrame:
    """Get failure stats from output of ``sim.accuracy.monte_carlo``.
    
    
    :param data: output from ``sim.accuracy.monte_carlo``.
    :param p_slice: slice of probabilities to restrict output to.
    :param alpha: significance level of confidence intervals.
    :param method: method to compute confidence intervals.
        For details on confidence intervals,
        see https://www.statsmodels.org/dev/generated/statsmodels.stats.proportion.proportion_confint.html.
    
    
    :return dT: a DataFrame where each row a (distance, probability); columns are:
    
        * ``f`` logical failure rate
        * ``lo`` lower confidence bound of ``f``
        * ``hi`` upper confidence bound of ``f``
        * ``x`` log10(``p``)
        * ``y`` log10(``f``)
        * ``yerr`` half the confidence interval of ``y``.
    """
    dT: DataFrame = data.loc[:, (slice(None), p_slice)].T # type: ignore
    dT['f'] = dT.m / dT.n
    dT['lo'], dT['hi'] = proportion_confint(dT.m, dT.n, alpha=alpha, method=method)
    dT['x'] = np.log10(dT.index.get_level_values('p'))
    dT['y']  = np.log10(dT.f)
    dT['yerr'] = (np.log10(dT.hi) - np.log10(dT.lo)) / 2
    del dT['m']
    del dT['n']
    return dT


def get_failure_data_from_subset_sample(
        data: DataFrame,
        code_class: Type[Code],
        noise: NoiseModel,
        noise_levels: npt.NDArray[np.float64],
        alpha: float = STANDARD_ERROR_ALPHA,
        method='normal',
):
    """Get failure stats from output of ``sim.accuracy.subset_sample``.
    
    :return: A DataFrame indexed by (distance, noise level), with columns:
    
        * ``f`` logical error probability.
        * ``lo`` lower bound of ``f``.
        * ``hi`` upper bound of ``f``.
    
    Side effects: adds columns ``f``, ``SE_lo``, ``SE_hi`` to ``data``.
    """
    data['f'] = data.m / data.n
    lo, hi = proportion_confint(data.m, data.n, alpha=alpha, method=method)
    data['SE_lo'] = data.f - lo
    data['SE_hi'] = hi - data.f
    data.fillna(0, inplace=True)
    ls = []
    for d, df in data.groupby(level='d'): # type: ignore
        d: int
        df = df.droplevel('d')
        code = code_class(d, noise=noise)
        f, lo, hi = [], [], []
        for noise_level in noise_levels:
            subset_probs = code.NOISE.subset_probabilities(noise_level, survival=False)
            considered = df.join(subset_probs, rsuffix=' current')
            ignored = subset_probs[~subset_probs.index.isin(df.index)]

            mean = np.dot(considered['subset prob current'], considered['f'])
            se_lo = np.linalg.norm(considered['subset prob current'] * considered['SE_lo'])
            se_hi = np.linalg.norm(considered['subset prob current'] * considered['SE_hi'])
            cutoff = ignored['subset prob'].sum()

            f.append(mean)
            lo.append(mean - se_lo)
            hi.append(mean + se_hi + cutoff)
        ls.append(DataFrame({
            'f': f,
            'lo': lo,
            'hi': hi,
        }, index=noise_levels))
    return pd.concat(
        ls,
        keys=data.index.get_level_values('d').unique(),
        names=['d', 'p'],
    )


def get_log_runtime_data(data: DataFrame):
    """Get log runtime data from runtime data.
    
    
    ``data`` output from ``sim.runtime.batch``.
    
    
    :returns log_data: A DataFrame where each column a stat:
        ``x``, ``y``, or ``yerr``; row, a (probability, distance).
    """
    df: DataFrame = np.log10(data) # type: ignore
    log_data = DataFrame({
        'x': np.log10(data.columns.get_level_values('d')),
        'y': df.mean(),
        'yerr': df.sem(),
    })
    return log_data.swaplevel(0, 1, axis=0).sort_index(axis=0)


def get_stats(log_data: DataFrame, missing='drop', **kwargs_for_WLS):
    """Get WLS (weighted least squares) stats from output of either
    ``get_log_runtime_data`` OR ``get_failure_data``.
    
    
    :param log_data: output from ``get_log_runtime_data`` or ``get_failure_data``.
    :param missing: passed to ``statsmodels.regression.linear_model.WLS``.
    :param kwargs_for_WLS: passed to ``statsmodels.regression.linear_model.WLS``.
    
    :return stats: A DataFrame where each row an x-value
        (probability OR distance);
        columns are:
        
        * ``intercept``
        * ``se_intercept``
        * ``gradient``
        * ``se_gradient``
        * ``r_squared``
    """
    records: list[tuple[
        int | float,
        float, float,
        float, float,
    ]] = []
    index_name: str
    index_name, _ = log_data.index.names # type: ignore
    for idx, df in log_data.groupby(level=0):
        df=df.droplevel(index_name, axis=0)
        results = sm.WLS(
            endog=df['y'],
            exog=sm.add_constant(df['x']),
            weights=1/df['yerr']**2, # type: ignore
            missing=missing,
            **kwargs_for_WLS,
        ).fit()
        intercept, gradient = results.params
        se_intercept, se_gradient = results.bse
        r_squared = results.rsquared
        records.append((
            idx, # type: ignore
            intercept, se_intercept,
            gradient, se_gradient,
            r_squared,
        ))
    stats = DataFrame.from_records(
        records,
        index=index_name,
        columns=[
            index_name,
            'intercept', 'se_intercept',
            'gradient', 'se_gradient',
            'r_squared',
        ]
    )
    return stats


def add_ignored_timesteps(
    data: DataFrame,
    extra_steps_per_layer=2,
    layers_per_sample: Callable[[int], int] = lambda d: d,
):
    """Add ignored timesteps to ``data``.
    
    
    :param data: a DataFrame where each
        column a (distance, noise level);
        row, a runtime sample.
    :param extra_steps_per_layer: number of ignored timesteps per layer
        in the decoding graph.
    :param layers_per_sample: a function with input ``d`` that outputs
        the measurement round count per sample.
    
    E.g. if ``data`` is output of ``sim.runtime.frugal`` with ``time_only='merging'``
    using Snowflake with the 1:1 schedule,
    the default kwargs of this function will convert it to
    the analogous output of ``sim.runtime.frugal`` with ``time_only='all'``.
    This is because the extra timesteps per layer are a drop and a grow.
    If using Snowflake with the 2:1 schedule,
    set ``extra_steps_per_layer=3`` as there is one additional grow timestep.
    """
    dc = {(d, noise_level): data[d, noise_level] + extra_steps_per_layer*layers_per_sample(int(d))
        for d, noise_level in data.columns}
    return DataFrame(dc, columns=data.columns)


from unittest import mock

from localuf.decoders.luf import ActisNode, ActisNodes, Waiter

def test_init(actis_nodes: ActisNodes):

    assert type(actis_nodes._dc) is dict
    assert len(actis_nodes._dc) == len(actis_nodes.LUF.CODE.NODES)
    v, node = next(iter(actis_nodes._dc.items()))
    assert type(v) is tuple
    assert type(node) is ActisNode

    d = actis_nodes.LUF.CODE.D
    noise = actis_nodes.LUF.CODE.NOISE
    if str(noise) == 'code capacity':
        assert actis_nodes._SPAN == 1 + d-1 + d
    elif str(noise) == 'phenomenological':
        assert actis_nodes._SPAN == 1 + d-1 + d + d-1
    elif str(noise) == 'circuit-level':
        assert actis_nodes._SPAN == 1 + d
    else:
        raise ValueError(f"Unknown noise model: {noise}")

    assert isinstance(actis_nodes._WAITER, Waiter)

    with (
        mock.patch("localuf.decoders.luf.Nodes.__init__") as mock_init,
        mock.patch("localuf.decoders.luf.ActisNodes.reset") as mock_reset,
    ):
        actis_nodes.__init__(actis_nodes.LUF)
        mock_init.assert_called_once_with(actis_nodes.LUF)
        mock_reset.assert_called_once_with()

def test_reset(actis5Fu_nodes: ActisNodes):
    actis5Fu_nodes.busy = True
    actis5Fu_nodes.valid = False
    actis5Fu_nodes.countdown = 1
    actis5Fu_nodes.busy_signal = True
    actis5Fu_nodes.next_busy_signal = True
    actis5Fu_nodes.active_signal = True
    actis5Fu_nodes.next_active_signal = True

    actis5Fu_nodes.reset()

    assert not actis5Fu_nodes.busy
    assert actis5Fu_nodes.valid
    assert actis5Fu_nodes.countdown == 0
    assert not actis5Fu_nodes.busy_signal
    assert not actis5Fu_nodes.next_busy_signal
    assert not actis5Fu_nodes.active_signal
    assert not actis5Fu_nodes.next_active_signal

def test_update_valid(actis_nodes: ActisNodes):
    assert actis_nodes.valid
    actis_nodes.update_valid()
    assert actis_nodes.valid
    actis_nodes.active_signal = True
    actis_nodes.update_valid()
    assert not actis_nodes.valid
    assert not actis_nodes.active_signal

def test_advance(actis_nodes: ActisNodes):
    with (
        mock.patch("localuf.decoders.luf.Nodes.advance") as mock_ufn,
        mock.patch("localuf.decoders.luf.OptimalWaiter.advance") as mock_ew,
    ):
        actis_nodes.advance()
        mock_ufn.assert_called_once_with()
        mock_ew.assert_called_once_with()


import copy
from unittest import mock

import numpy as np
import pytest

from localuf.noise import CircuitLevel
from localuf.type_aliases import EdgeType, Edge, MultiplicityVector

@pytest.fixture(name="cl", params=[
    'standard',
    'balanced',
    'ion trap',
])
def _cl(request):
    return CircuitLevel(
        fresh_edge_dict={'S': (((0, 0, 0), (1, 0, 0)),)},
        parametrization=request.param,
        demolition=False,
        monolingual=False,
    )

def test_DEFAULT_MULTIPLICITIES():
    assert type(CircuitLevel._DEFAULT_MULTIPLICITIES) is dict
    assert len(CircuitLevel._DEFAULT_MULTIPLICITIES) == 12
    assert len(CircuitLevel._DEFAULT_MULTIPLICITIES['S']) == 4

def test_ALL_COEFFICIENTS():
    assert type(CircuitLevel._ALL_COEFFICIENTS) is dict
    assert len(CircuitLevel._ALL_COEFFICIENTS) == 3
    assert len(CircuitLevel._ALL_COEFFICIENTS['standard']) == 4

@pytest.mark.parametrize("parametrization", [
    'standard',
    'balanced',
    'ion trap',
])
def test_init(parametrization):
    edge_dict: dict[EdgeType, tuple[Edge, ...]] = {'S': (((0, 0, 0), (1, 0, 0)),)}
    with mock.patch(
        "localuf.noise.CircuitLevel._make_multiplicities",
        return_value=CircuitLevel._make_multiplicities(demolition=False, monolingual=False),
    ) as mock_gm:
        cl = CircuitLevel(
            fresh_edge_dict=edge_dict,
            parametrization=parametrization,
            demolition=False,
            monolingual=False,
        )
        mock_gm.assert_called_once_with(False, False)
    assert cl._COEFFICIENTS == CircuitLevel._ALL_COEFFICIENTS[parametrization]


def test_get_dc(
        toy_edge_dict: dict[EdgeType, tuple[Edge, ...]],
        toy_merges: dict[Edge, Edge],
):
    """`dc` should have two entries whose values are different."""
    e0, e1 = toy_edge_dict['E westmost']
    toy_multiplicities: dict[EdgeType, MultiplicityVector] = {
        et: np.array(m) for et, m
        in CircuitLevel._DEFAULT_MULTIPLICITIES.items()
    }
    dc = CircuitLevel._get_dc(
        toy_edge_dict,
        toy_multiplicities,
        toy_merges,
    )
    assert len(dc) == 2
    assert (dc[e0] == toy_multiplicities['E westmost']).all()
    assert (dc[e1] == sum((
            toy_multiplicities['E westmost'],
            toy_multiplicities['EU west corners'],
            toy_multiplicities['SEU'],
        ))).all()


def test_make_multiplicities(assert_these_multiplicities_unchanged):
    m = CircuitLevel._make_multiplicities(demolition=False, monolingual=False)
    assert_these_multiplicities_unchanged(m, CircuitLevel._DEFAULT_MULTIPLICITIES.keys())

def test_make_multiplicities_demolition(
        assert_these_multiplicities_unchanged,
        diagonals,
):
    m = CircuitLevel._make_multiplicities(demolition=True, monolingual=False)
    assert (m['S'] == np.array((4, 2, 2, 0))).all()
    assert (m['E westmost'] == np.array((1, 0, 3, 0))).all()
    assert (m['E bulk'] == np.array((2, 0, 2, 0))).all()
    assert (m['E eastmost'] == np.array((1, 0, 2, 0))).all()
    assert (m['U 3'] == np.array((3, 0, 1, 2))).all()
    assert (m['U 4'] == np.array((4, 0, 0, 2))).all()
    assert_these_multiplicities_unchanged(m, diagonals)

def test_make_multiplicities_monolingual(
        assert_these_multiplicities_unchanged,
        diagonals,
):
    m = CircuitLevel._make_multiplicities(demolition=False, monolingual=True)
    assert (m['S'] == np.array((4, 2, 3, 0))).all()
    assert (m['E westmost'] == np.array((1, 0, 4, 0))).all()
    assert (m['E bulk'] == np.array((2, 0, 3, 0))).all()
    assert (m['E eastmost'] == np.array((1, 0, 3, 0))).all()
    assert (m['U 3'] == np.array((3, 0, 3, 1))).all()
    assert (m['U 4'] == np.array((4, 0, 2, 1))).all()
    assert_these_multiplicities_unchanged(m, diagonals)

def test_DEFAULT_MULTIPLICITIES_unchanged(assert_these_multiplicities_unchanged):
    dc = copy.deepcopy(CircuitLevel._DEFAULT_MULTIPLICITIES)
    m = CircuitLevel._make_multiplicities(demolition=True, monolingual=True)
    assert_these_multiplicities_unchanged(dc, dc.keys())
    m['S'] += np.array((1, 1, 1, 1))
    assert_these_multiplicities_unchanged(dc, ('S',))


def test_make_error(toy_cl: CircuitLevel, e_westmost: tuple[Edge, Edge]):
    m0, m1 = toy_cl._FRESH_EDGES.keys()
    for noise_level in (0, 1):
        with mock.patch(
            'localuf.noise.CircuitLevel._get_flip_probabilities',
            return_value={m0: 1-noise_level, m1: noise_level},
        ) as mock_gep:
            error = toy_cl.make_error(noise_level)
            mock_gep.assert_called_once_with(noise_level)
            assert error == {e_westmost[noise_level]}


def test_get__flip_probabilities(cl: CircuitLevel):
    with mock.patch('localuf.noise._multiset_handler.MultisetHandler.pr') as mock_pr:
        cl._get_flip_probabilities(0)
        cl._get_flip_probabilities(1)
        assert mock_pr.call_args_list == [
            mock.call(m, (0, 0, 0, 0))
            for m in cl._FRESH_EDGES.keys()
        ] + [
            mock.call(m, cl._COEFFICIENTS)
            for m in cl._FRESH_EDGES.keys()
        ]


def test_get_edge_weights(cl: CircuitLevel):
    noise_level = 1e-1
    with mock.patch("localuf.noise.CircuitLevel._get_flip_probabilities") as mock_gfp:
        cl.get_edge_weights(noise_level)
        mock_gfp.assert_called_once_with(noise_level)


from unittest import mock

import pytest

from localuf.type_aliases import Node
from localuf.constants import Growth
from localuf.decoders import Snowflake
from localuf.decoders.policies import DecodeDrawer

@pytest.mark.parametrize("unrooter", ['full', 'simple'])
def test_init(snowflake: Snowflake, unrooter):
    with mock.patch("localuf.decoders.uf.BaseUF.__init__") as mock_init:
        snowflake.__init__(snowflake.CODE, unrooter=unrooter)
        mock_init.assert_called_once_with(snowflake.CODE)
    assert type(snowflake._NODES) is dict
    assert type(snowflake._EDGES) is dict
    assert set(snowflake._NODES.keys()) == {
        v for v in snowflake.CODE.NODES
        if v[snowflake.CODE.TIME_AXIS] < snowflake.CODE.SCHEME.WINDOW_HEIGHT
    }
    assert set(snowflake._EDGES.keys()) == {
        e for e in snowflake.CODE.EDGES
        if e[1][snowflake.CODE.TIME_AXIS] < snowflake.CODE.SCHEME.WINDOW_HEIGHT
    }
    assert type(snowflake._DECODE_DRAWER) is DecodeDrawer
    assert snowflake._DECODE_DRAWER._FIG_WIDTH == snowflake._FIG_WIDTH
    assert snowflake._DECODE_DRAWER._FIG_HEIGHT == snowflake._FIG_HEIGHT

@pytest.mark.parametrize("prop", [
    "NODES",
    "EDGES",
    "syndrome",
    "growth",
    "correction",
    "_pointer_digraph",
    "_FIG_WIDTH",
    "_FIG_HEIGHT",
])
def test_property_attributes(test_property, snowflake: Snowflake, prop):
    test_property(snowflake, prop)


def test_syndrome(snowflake: Snowflake,):
    snowflake.NODES[0, 0].defect = True
    assert snowflake.syndrome == {(0, 0)}
    snowflake.NODES[-1, 0].defect = True
    assert snowflake.syndrome == {(0, 0)}


@pytest.mark.parametrize("v", [(0, 0), (-1, 0)])
def test_verbose_syndrome(snowflake: Snowflake, v: Node):
    snowflake.NODES[v].defect = True
    assert snowflake.verbose_syndrome == {v}


def test_growth(snowflake: Snowflake, uvw: tuple[Node, Node, Node]):
    u, v, w = uvw
    snowflake.EDGES[u, v].growth = Growth.HALF
    snowflake.EDGES[v, w].growth = Growth.FULL
    for e in snowflake.EDGES.keys():
        if e == (u, v):
            assert snowflake.growth[e] is Growth.HALF
        elif e == (v, w):
            assert snowflake.growth[e] is Growth.FULL
        else:
            assert snowflake.growth[e] is Growth.UNGROWN


@pytest.fixture
def index_to_id_helper():
    def f(uf: Snowflake):
        d = uf.CODE.D
        boundary_IDs: set[int] = set()
        detector_IDs: set[int] = set()
        for v in uf.NODES.keys():
            id_ = uf.index_to_id(v)
            if uf.CODE.is_boundary(v):
                boundary_IDs.add(id_)
            else:
                detector_IDs.add(id_)
        assert max(boundary_IDs) < min(detector_IDs)
        return d, boundary_IDs, detector_IDs
    return f


def test_index_to_id_2D(snowflake: Snowflake, index_to_id_helper):
    h = snowflake.CODE.SCHEME.WINDOW_HEIGHT
    d, boundary_IDs, detector_IDs = index_to_id_helper(snowflake)
    assert len(boundary_IDs) + len(detector_IDs) == h * (d+1)
    # highest ID boundary node
    assert snowflake.index_to_id((d-1, 0)) == 2*h-1
    # highest ID detector
    assert snowflake.index_to_id((d-2, 0)) == h * (d+1) - 1


def test_reset(snowflake: Snowflake):
    snowflake.history = [snowflake]
    with (
        mock.patch("localuf.decoders.uf.BaseUF.reset") as uf_reset,
        mock.patch("localuf.decoders.snowflake._Node.reset") as node_reset,
        mock.patch("localuf.decoders.snowflake._Edge.reset") as edge_reset,
    ):
        snowflake.reset()
        uf_reset.assert_called_once_with()
        assert node_reset.call_args_list == [mock.call()] * len(snowflake.NODES)
        assert edge_reset.call_args_list == [mock.call()] * len(snowflake.EDGES)
    assert not hasattr(snowflake, 'history')


def test_init_history(snowflake3: Snowflake):
    assert not hasattr(snowflake3, 'history')
    snowflake3.init_history()
    assert snowflake3.history == []


tds = "localuf.decoders.snowflake"


class TestDecode:

    syndrome: set[Node] = {(0, 0)}
    time_only = 'merging'

    def two_one(self, snowflake3: Snowflake):
        log_history = 'fine'
        return_value = 1
        with (
            mock.patch(f"{tds}.Snowflake.append_history") as mock_ah,
            mock.patch(f"{tds}.Snowflake.drop") as mock_drop,
            mock.patch(
                f"{tds}.Snowflake.merge",
                return_value=return_value,
            ) as mock_merge,
        ):
            assert snowflake3.decode(self.syndrome, log_history=log_history) == 2*return_value
            assert mock_ah.call_args_list == [mock.call()] * 3
            mock_drop.assert_called_once_with(self.syndrome)
            assert mock_merge.call_args_list == [
                mock.call(True, log_history, time_only=self.time_only),
                mock.call(False, log_history, time_only=self.time_only),
            ]

    def one_one(self, snowflake3_one_one: Snowflake):
        return_value = 'return_value'
        log_history = 'fine'
        with (
            mock.patch(f"{tds}.Snowflake.append_history") as mock_ah,
            mock.patch(f"{tds}.Snowflake.drop") as mock_drop,
            mock.patch(
                f"{tds}.Snowflake.merge",
                return_value=return_value,
            ) as mock_merge,
        ):
            assert snowflake3_one_one.decode(self.syndrome, log_history=log_history) == return_value
            assert mock_ah.call_args_list == [mock.call()] * 2
            mock_drop.assert_called_once_with(self.syndrome)
            mock_merge.assert_called_once_with(
                whole=True,
                log_history=log_history,
                time_only=self.time_only,
            )


def test_drop(snowflake: Snowflake):
    h = snowflake.CODE.SCHEME.WINDOW_HEIGHT
    syndrome = {
        (0, h-1),
        (2, h-1),
    }
    with (
        mock.patch(f"{tds}.EdgeContact.drop") as ec,
        mock.patch(f"{tds}.FloorContact.drop") as gc,
        mock.patch(f"{tds}._Edge.update_after_drop") as euu,
        mock.patch(f"{tds}.NodeFriendship.drop") as node_f,
        mock.patch(f"{tds}.NothingFriendship.drop") as nothing_f,
        mock.patch(f"{tds}._Node.update_after_drop") as nuu,
        mock.patch(f"{tds}.Snowflake._load") as mock_load,
    ):
        snowflake.drop(syndrome)
        assert ec.call_args_list \
            + gc.call_args_list == [mock.call()] * len(snowflake.EDGES)
        assert euu.call_args_list == [mock.call()] * len(snowflake.EDGES)
        assert node_f.call_args_list \
            + nothing_f.call_args_list == [mock.call()] * len(snowflake.NODES)
        assert nuu.call_args_list == [mock.call()] * len(snowflake.NODES)
        mock_load.assert_called_once_with(syndrome)


def test_load(snowflake3: Snowflake):
    h = snowflake3.CODE.SCHEME.WINDOW_HEIGHT
    nodes = (
        (0, h-1),
        (2, h-1),
    )
    snowflake3.NODES[nodes[0]].next_defect = True
    snowflake3._load(set(nodes))
    for v, node in snowflake3.NODES.items():
        assert node.next_defect is (v==nodes[1])


def test_merge(snowflake: Snowflake):
    whole = True
    node_count = len(snowflake.NODES)
    t = 0

    with (
        mock.patch(f"{tds}.Snowflake.append_history") as mock_ah,
        mock.patch(f"{tds}._Node.merging") as mock_merging,
        mock.patch(f"{tds}._Node.update_after_merging") as mock_uam,
    ):
        assert snowflake.merge(whole, 'fine') == t
        mock_ah.assert_not_called()
        assert mock_merging.call_args_list == node_count * [mock.call(whole)]
        assert mock_uam.call_args_list == node_count * [mock.call()]


def test_append_history(snowflake3: Snowflake):
    snowflake3.history = []
    snowflake3.append_history()
    assert type(snowflake3.history) is list
    assert len(snowflake3.history) == 1
    assert type(snowflake3.history[0]) is Snowflake


"""Module for CSS codes.

Available codes:

* Repetition
* Surface
"""

from collections.abc import Iterable
from functools import cache
from itertools import repeat, chain

from localuf import constants
from localuf.type_aliases import Coord, Edge, EdgeType, Node
from localuf._base_classes import Code
from localuf._schemes import Batch


class Repetition(Code):
    """The decoding graph of a repetition code.
    
    Extends ``Code`` class.
    """

    _LONG_AXIS = 0
    _CODE_DIMENSION = 1

    def _code_capacity_edges(
            self,
            merge_equivalent_boundary_nodes: bool,
    ) -> tuple[int, tuple[Edge, ...]]:
        d = self.D
        return d, tuple(((j,), (j+1,)) for j in range(-1, d-1))

    def _phenomenological_edges(
            self,
            h,
            future_boundary,
            t_start=0,
            merge_equivalent_boundary_nodes=False,
    ) -> tuple[int, tuple[Edge, ...]]:
        d = self.D
        layer_count = h if future_boundary else h-1
        n_edges = h*d + layer_count*(d-1)
        if merge_equivalent_boundary_nodes:
            j_edges = [((-1, 0), (0, t)) for t in range(t_start, t_start+h)] \
            + [((j, t), (j+1, t)) for j in range(d-2) for t in range(t_start, t_start+h)] \
            + [((d-2, t), (d-1, 0)) for t in range(t_start, t_start+h)]
        else:
            j_edges = (((j, t), (j+1, t)) for j in range(-1, d-1) for t in range(t_start, t_start+h))
        t_edges = (((j, t), (j, t+1)) for j in range(d-1) for t in range(t_start, t_start+layer_count))
        return n_edges, (*j_edges, *t_edges)

    def _future_boundary_nodes(self, h) -> list[Node]:
        return [(j, h) for j in range(self.D-1)]
    
    def _redundant_boundary_nodes(self, h):
        raise NotImplementedError("Yet to implement circuit-level noise for repetition code.")
    
    def _circuit_level_edges(self, **_):
        raise NotImplementedError("Yet to implement circuit-level noise for repetition code.")

    @cache
    def index_to_id(self, index: Node):
        """Return ID of node at index (j,) or (j, t)."""
        d = self.D
        h = self.SCHEME.WINDOW_HEIGHT
        if len(index) == 2:
            j, t = index
            if j == -1:  # on west boundary
                return 2*t
            elif j == d-1:  # on east boundary
                return 2*t + 1
            elif t == h:  # on future boundary
                spatial_boundary_count = 2*h
                return spatial_boundary_count + j
            else:  # not a boundary
                boundary_count = 2*h if isinstance(self.SCHEME, Batch) else 2*h + d-1
                return boundary_count + (d-1)*t + j
        else:  # 1
            j, = index
            return j+1

    def get_pos(
            self,
            x_offset: float = constants.DEFAULT_X_OFFSET,
            nodelist: None | Iterable[Node] = None,
    ) -> dict[Node, Coord]:
        if nodelist is None:
            nodelist = self.NODES
        if self.DIMENSION == 1:
            pos = {(j,): (j, 0) for j, in nodelist}
        else:
            pos = dict(zip(nodelist, nodelist))
        return pos


class Surface(Code):
    """The decoding graph of an unrotated surface code with boundaries.
    
    Extends ``Code`` class.
    
    Additional instance properties:
    
    * ``DATA_QUBIT_COUNT`` number of data qubits.
    """

    _LONG_AXIS = 1
    _CODE_DIMENSION = 2

    @property
    def DATA_QUBIT_COUNT(self) -> int:
        d = self.D
        return d**2 + (d-1)**2

    def _code_capacity_edges(
            self,
            merge_equivalent_boundary_nodes: bool,
    ) -> tuple[int, tuple[Edge, ...]]:
        d = self.D
        i_edges = (((i, j), (i+1, j)) for i in range(d-1) for j in range(    d-1))
        if merge_equivalent_boundary_nodes:
            j_edges = [((0, -1), (i, 0)) for i in range(d)] \
                + [((i, j), (i, j+1)) for i in range(d  ) for j in range(d-2)] \
                + [((i, d-2), (0, d-1)) for i in range(d)]
        else:
            j_edges = (((i, j), (i, j+1)) for i in range(d  ) for j in range(-1, d-1))
        return self.DATA_QUBIT_COUNT, (*i_edges, *j_edges)

    def _phenomenological_edges(
            self,
            h,
            future_boundary,
            t_start=0,
            merge_equivalent_boundary_nodes=False,
    ) -> tuple[int, tuple[Edge, ...]]:
        d = self.D
        layer_count = h if future_boundary else h-1
        n_edges = h * self.DATA_QUBIT_COUNT + layer_count * d*(d-1)
        i_edges = (((i, j, t), (i+1, j, t)) for i in range(d-1) for j in range(    d-1) for t in range(t_start, t_start+h))
        if merge_equivalent_boundary_nodes:
            j_edges = [((0, -1, 0), (i, 0, t)) for i in range(d) for t in range(t_start, t_start+h)] \
                + [((i, j, t), (i, j+1, t)) for i in range(d  ) for j in range(d-2) for t in range(t_start, t_start+h)] \
                + [((i, d-2, t), (0, d-1, 0)) for i in range(d) for t in range(t_start, t_start+h)]
        else:
            j_edges = (((i, j, t), (i, j+1, t)) for i in range(d  ) for j in range(-1, d-1) for t in range(t_start, t_start+h))
        t_edges = (((i, j, t), (i, j, t+1)) for i in range(d  ) for j in range(    d-1) for t in range(t_start, t_start+layer_count))
        return n_edges, (*i_edges, *j_edges, *t_edges)

    def _future_boundary_nodes(self, h) -> list[Node]:
        d = self.D
        return [(i, j, h) for i in range(d) for j in range(d-1)]
    
    def _redundant_boundary_nodes(self, h) -> list[Node]:
        d = self.D
        return [(i, j, t) for i in range(d) for j, t in ((-1, -1), (d-1, h))]

    def _circuit_level_edges(
            self,
            h: int,
            future_boundary: bool,
            _merge_redundant_edges: bool,
            t_start=0,
            merge_equivalent_boundary_nodes=False,
    ) -> tuple[
        int,
        tuple[Edge, ...],
        dict[EdgeType, tuple[Edge, ...]],
        dict[Edge, Edge] | None,
    ]:
        d = self.D
        j_ranges = ((-1,), range(d-2), (d-2,))
        if future_boundary:
            layer_count = h
            t_ranges = (
                range(t_start-1, t_start-1+layer_count),
                *repeat(range(t_start, t_start+layer_count), 2)
            )
        else:
            layer_count = h-1
            t_ranges = tuple(repeat(range(t_start, t_start+layer_count), 3))
        # south
        s = tuple(((i, j, t), (i+1, j, t)) for i in range(d-1) for j in range(d-1) for t in range(t_start, t_start+h))
        # east [westmost, bulk, eastmost]
        e_wm, e_bulk, e_em = (tuple(
            ((i, j, t), (i, j+1, t)) for i in range(d) for j in js for t in range(t_start, t_start+h)
        ) for js in j_ranges)
        if merge_equivalent_boundary_nodes:
            e_wm = tuple(((0, -1, 0), (i, 0, t)) for i in range(d) for t in range(t_start, t_start+h))
            e_em = tuple(((i, d-2, t), (0, d-1, 0)) for i in range(d) for t in range(t_start, t_start+h))
        # north- or southmost up, up
        u3, u4 = (tuple(
            ((i, j, t), (i, j, t+1)) for i in is_ for j in range(d-1) for t in range(t_start, t_start+layer_count)
        ) for is_ in ((0, d-1), range(1, d-1)))
        # south down
        sd = tuple(((i, j, t), (i+1, j, t-1)) for i in range(d-1) for j in range(d-1) for t in range(t_start+1, t_start+1+layer_count))
        # east up [west corners, on north or south wall, east corners]
        eu_wc, eu_edge_NS, eu_ec = (tuple(
            ((i, j, t), (i, j+1, t+1)) for i in (0, d-1) for j in js for t in ts
        ) for js, ts in zip(j_ranges, t_ranges))
        # east up [on west wall, in centre, on east wall]
        eu_edge_W, eu_centre, eu_edge_E = (tuple(
                ((i, j, t), (i, j+1, t+1)) for i in range(1, d-1) for j in js for t in ts
        ) for js, ts in zip(j_ranges, t_ranges))
        # east up on east or west wall
        eu_edge_EW = (*eu_edge_W, *eu_edge_E)
        # east up on a wall
        eu_edge = (*eu_edge_NS, *eu_edge_EW)
        # south east up [westmost, bulk, eastmost]
        seu_W, seu_bulk, seu_E = (tuple(
                ((i, j, t), (i+1, j+1, t+1)) for i in range(d-1) for j in js for t in ts
        ) for js, ts in zip(j_ranges, t_ranges))
        # south east up west- or eastmost
        seu_boundary = (*seu_W, *seu_E)
        # south east up
        seu = (*seu_boundary, *seu_bulk)
        if _merge_redundant_edges or merge_equivalent_boundary_nodes:
            n_edges = h * self.DATA_QUBIT_COUNT + layer_count * (2*d-1)*(2*d-3)
            edges = (
                *s,
                *e_wm, *e_bulk, *e_em,
                *u3, *u4,
                *sd,
                *eu_edge_NS, *eu_centre,
                *seu_bulk,
            )
            merges = {e: self._substitute(e, merge_equivalent_boundary_nodes)
                for e in (*eu_wc, *eu_ec, *eu_edge_EW, *seu_boundary)}
        else:
            n_edges = (h+layer_count) * self.DATA_QUBIT_COUNT + 2 * d * (d-1) * layer_count
            edges = (
                *s,
                *e_wm, *e_bulk, *e_em,
                *u3, *u4,
                *sd,
                *eu_wc, *eu_ec, *eu_edge, *eu_centre,
                *seu,
            )
            merges = None
        edge_dict: dict[EdgeType, tuple[Edge, ...]] = {
            'S': s,
            'E westmost': e_wm,
            'E bulk': e_bulk,
            'E eastmost': e_em,
            'U 3': u3,
            'U 4': u4,
            'SD': sd,
            'EU west corners': eu_wc,
            'EU east corners': eu_ec,
            'EU edge': eu_edge,
            'EU centre': eu_centre,
            'SEU': seu,
        }
        return n_edges, edges, edge_dict, merges

    def _substitute(self, e: Edge, merge_equivalent_boundary_nodes: bool = False) -> Edge:
        """Return substitute edge for redundant edge ``e``."""
        d = self.D
        a = self.LONG_AXIS
        u, v = e
        remaining_dimensions = len(u) - a - 1
        if u[a] == -1:
            w = tuple(chain(repeat(0, a), (-1,), repeat(0, remaining_dimensions))
            ) if merge_equivalent_boundary_nodes else ((*v[:a], -1, *v[a+1:]))
            return (w, v)
        elif v[a] == d-1:
            w = tuple(chain(repeat(0, a), (d-1,), repeat(0, remaining_dimensions))
            ) if merge_equivalent_boundary_nodes else (*u[:a], d-1, *u[a+1:])
            return (u, w)
        else:
            raise ValueError(f'Edge {e} must have a boundary node.')

    def index_to_label(self, index: Node):
        """Return node label of measure-Z qubit at index (i, j)."""
        if self.DIMENSION == 2:
            i, j = index
            return (self.D+1)*i + j + 1
        else:
            raise NotImplementedError('Only implemented for code capacity.')

    def label_to_index(self, a: int) -> Node:
        """Return measure-Z qubit index (i, j) of node label a."""
        d = self.D
        if self.DIMENSION == 2:
            return (a // (d+1), a % (d+1) - 1)
        else:
            raise NotImplementedError('Only implemented for code capacity.')

    @cache
    def index_to_id(self, index: Node):
        """Return ID of node at index (i, j) or (i, j, t)."""
        d = self.D
        if len(index) == 2:
            i, j = index
            if j == -1:  # on west boundary
                return 2*i
            elif j == d-1:  # on east boundary
                return 2*i + 1
            else:  # not a boundary
                boundary_count = 2 * d
                return boundary_count + (d-1)*i + j
        else:  # 3
            h = self.SCHEME.WINDOW_HEIGHT
            i, j, t = index
            if j == -1:  # on west boundary
                return 2*h*i + t
            elif j == d-1:  # on east boundary
                return (2*i + 1) * h + t
            elif t == h:  # on future boundary
                spatial_boundary_count = 2*d*h
                return spatial_boundary_count + (d-1)*i + j
            else:  # not a boundary
                boundary_count = 2*d*h if isinstance(self.SCHEME, Batch) else 2*d*h + d*(d-1)
                return boundary_count + h*(d-1)*i + h*j + t

    def get_pos(
            self,
            x_offset: float = constants.DEFAULT_X_OFFSET,
            nodelist: None | Iterable[Node] = None,
    ) -> dict[Node, Coord]:
        if nodelist is None:
            nodelist = self.NODES
        if self.DIMENSION == 2:
            pos = {(i, j): (j, -i) for i, j in nodelist}
        else:
            pos = {(i, j, t): (j+x_offset*i, t-x_offset*i)
                for i, j, t in nodelist}
        return pos


from unittest import mock

import pytest

from localuf import Repetition, Surface
from localuf._determinants import Determinant
from localuf._schemes import Batch, Forward


@pytest.mark.parametrize("prop", [
    "WINDOW_HEIGHT",
])
def test_property_attributes(test_property, batch3F: Batch, prop: str):
    test_property(batch3F, prop)

def test_CODE_attribute(sf3F: Surface, batch3F: Batch):
    assert batch3F._CODE is sf3F

def test_WINDOW_HEIGHT_attribute(forward_rp: Forward):
    assert forward_rp.WINDOW_HEIGHT == forward_rp._COMMIT_HEIGHT + forward_rp._BUFFER_HEIGHT

def test_DETERMINANT_attribute(batch3F: Batch):
    assert issubclass(type(batch3F._DETERMINANT), Determinant)

def test_get_logical_error(sf5F: Surface):
    leftover0 = {
        ((2, 2), (2, 3)),
        ((2, 3), (2, 4)),
        ((0, 2), (1, 2)),
        ((1, 2), (2, 2)),
        ((0, 2), (0, 3)),
        ((0, 3), (0, 4)),
    }
    leftover1 = {((0, j), (0, j+1)) for j in range(-1, 5-1)}
    assert sf5F.SCHEME.get_logical_error(leftover0) == 0
    assert sf5F.SCHEME.get_logical_error(leftover1) == 1


def test_is_boundary(sf3F: Surface, rp3_forward: Repetition):
    with mock.patch("localuf._determinants.Determinant.is_boundary") as m:
        v, *_ = sf3F.NODES
        sf3F.SCHEME.is_boundary(v)
        m.assert_called_once_with(v)
    with mock.patch("localuf._determinants.SpaceTimeDeterminant.is_boundary") as m:
        v, *_ = rp3_forward.NODES
        rp3_forward.SCHEME.is_boundary(v)
        m.assert_called_once_with(v)


from unittest import mock

import networkx as nx
import pytest

from localuf import constants, Surface, Repetition
from localuf._schemes import Scheme
from localuf.noise import CodeCapacity
from localuf.noise.main import Phenomenological
from localuf.type_aliases import Edge
from localuf._base_classes import Code

@pytest.fixture
def error_2D():
    return {
        ((0, -1), (0, 0)),
        ((0, 0), (0, 1)),
    }

@pytest.fixture
def error_3D():
    return {
        ((0, -1, 1), (0, 0, 1)),
        ((0, 0, 1), (0, 1, 1)),
    }

@pytest.mark.parametrize("prop", [
    "D",
    "SCHEME",
    "N_EDGES",
    "EDGES",
    "NODES",
    "TIME_AXIS",
    "LONG_AXIS",
    "INCIDENT_EDGES",
])
def test_property_attributes(test_property, sf3F: Surface, prop: str):
    test_property(sf3F, prop)


class ConcreteCode(Code):

    _EDGES = ()
    _N_EDGES = 0
    _NODES = ()

    def _code_capacity_edges(self, merge_equivalent_boundary_nodes):
        return super()._code_capacity_edges(merge_equivalent_boundary_nodes)

    def _phenomenological_edges(
            self,
            h,
            future_boundary,
            merge_equivalent_boundary_nodes,
    ):
        return super()._phenomenological_edges(
            h,
            future_boundary,
            merge_equivalent_boundary_nodes,
        )

    def _circuit_level_edges(
            self,
            h,
            future_boundary,
            _merge_redundant_edges,
            merge_equivalent_boundary_nodes,
    ): pass

    def _future_boundary_nodes(self, h): pass

    def _redundant_boundary_nodes(self, h): pass

    def index_to_id(self, index): pass

    def get_pos(self, x_offset, nodelist): pass


@pytest.mark.parametrize("noise", ('code capacity', 'phenomenological'))
@pytest.mark.parametrize("scheme", ('batch', 'global batch'))
def test_init(noise, scheme):
    with (
        mock.patch(
            "localuf._base_classes.Code._code_capacity_edges",
            return_value=(0, (((0, 0), (0, 1)),))
        ) as mock_code_capacity_edges,
        mock.patch(
            "localuf._base_classes.Code._phenomenological_edges",
            return_value=(0, (((0, 0, 0), (0, 1, 0)),))
        ) as mock_phenomenological_edges,
        mock.patch(
            "localuf._schemes.Batch.__init__",
            return_value=None
        ) as mock_batch_init,
        mock.patch(
            "localuf._schemes.Global.__init__",
            return_value=None
        ) as mock_global_batch_init,
    ):
        ConcreteCode(3, noise, scheme)
        if noise == 'phenomenological':
            mock_phenomenological_edges.assert_called_once()
            mock_code_capacity_edges.assert_not_called()
        else:
            mock_phenomenological_edges.assert_not_called()
            mock_code_capacity_edges.assert_called_once()
        if scheme == 'batch':
            mock_batch_init.assert_called_once()
            mock_global_batch_init.assert_not_called()
        else:
            mock_batch_init.assert_not_called()
            mock_global_batch_init.assert_called_once()

def test_D_attribute(sf: Surface):
    assert type(sf.D) is int


def test_SCHEME_attribute(sf: Surface):
    assert issubclass(type(sf.SCHEME), Scheme)


def test_INCIDENT_EDGES_attribute(sf: Surface):
    assert type(sf.INCIDENT_EDGES) is dict
    assert len(sf.INCIDENT_EDGES) == len(sf.NODES)


def test_NOISE_attribute(sf3F: Surface, sf3T: Surface):
    assert type(sf3F.NOISE) is CodeCapacity
    assert type(sf3T.NOISE) is Phenomenological


def test_INCIDENT_EDGES(sf):
    """Test `INCIDENT_EDGES` consistent with the old, slower calculation of incident edges."""
    assert sf.INCIDENT_EDGES == {v: {
        e for e in sf.EDGES if v in e
    } for v in sf.NODES}


def test_INCIDENT_EDGES_code_capacity(sf5F: Surface):
    # boundary node
    assert sf5F.INCIDENT_EDGES[0, -1] == {((0, -1), (0, 0))}
    # weight-3 stabilizer
    assert sf5F.INCIDENT_EDGES[0, 1] == {
        ((0, 0), (0, 1)),
        ((0, 1), (0, 2)),
        ((0, 1), (1, 1))
    }
    # weight-4 stabilizer
    assert sf5F.INCIDENT_EDGES[1, 1] == {
        ((0, 1), (1, 1)),
        ((1, 0), (1, 1)),
        ((1, 1), (1, 2)),
        ((1, 1), (2, 1))
    }


def test_INCIDENT_EDGES_phenomenological(sf5T: Surface):
    # boundary node
    assert sf5T.INCIDENT_EDGES[0, -1, 0] == {((0, -1, 0), (0, 0, 0))}
    # degree-5 node
    assert sf5T.INCIDENT_EDGES[0, 1, 1] == {
        ((0, 1, 0), (0, 1, 1)),
        ((0, 0, 1), (0, 1, 1)),
        ((0, 1, 1), (0, 2, 1)),
        ((0, 1, 1), (1, 1, 1)),
        ((0, 1, 1), (0, 1, 2))
    }
    # degree-6 node
    assert sf5T.INCIDENT_EDGES[1, 1, 1] == {
        ((1, 1, 0), (1, 1, 1)),
        ((1, 0, 1), (1, 1, 1)),
        ((1, 1, 1), (1, 2, 1)),
        ((0, 1, 1), (1, 1, 1)),
        ((1, 1, 1), (2, 1, 1)),
        ((1, 1, 1), (1, 1, 2))
    }


def test_is_boundary(sf3F: Surface):
    with mock.patch("localuf._base_classes.Scheme.is_boundary") as m:
        v, *_ = sf3F.NODES
        sf3F.is_boundary(v)
        m.assert_called_once_with(v)


def test_neighbors(sf5F: Surface, sf5T: Surface):
    assert sf5F.neighbors((0, -1)) == {(0, 0)}  # boundary node
    assert sf5F.neighbors((0, 1)) == {(0, 0), (0, 2), (1, 1)}  # weight-3 stabilizer
    assert sf5F.neighbors((1, 1)) == {(0, 1), (1, 0), (1, 2), (2, 1)}  # weight-4 stabilizer
    assert sf5T.neighbors((0, -1, 0)) == {(0, 0, 0)}
    assert sf5T.neighbors((0, 1, 1)) == {
        (0, 1, 0),
        (0, 0, 1),
        (0, 2, 1),
        (1, 1, 1),
        (0, 1, 2)
    }
    assert sf5T.neighbors((1, 1, 1)) == {
        (1, 1, 0),
        (0, 1, 1),
        (1, 0, 1),
        (1, 2, 1),
        (2, 1, 1),
        (1, 1, 2)
    }


def test_traverse_edge(sf3F: Surface):
    u, v = (0, 0), (0, 1)
    assert sf3F.traverse_edge((u, v), u) == v
    assert sf3F.traverse_edge((u, v), v) == u
    u, v = (0, 0, 0), (1, 0, 0)
    assert sf3F.traverse_edge((u, v), u) == v
    assert sf3F.traverse_edge((u, v), v) == u


def test_raise_node(sf3T: Surface):
    for v in sf3T.NODES:
        di, dj, dt = sf3T.raise_node(v)
        assert v == (di, dj, dt-1)
        li, lj, lt = sf3T.raise_node(v, -3)
        assert v == (li, lj, lt+3)


def test_raise_edge(sf3T: Surface):
    for e in sf3T.EDGES:
        (*uij, ut), (*vij, vt) = sf3T.raise_edge(e)
        assert e == ((*uij, ut-1), (*vij, vt-1))
        (*uij, ut), (*vij, vt) = sf3T.raise_edge(e, -3)
        assert e == ((*uij, ut+3), (*vij, vt+3))


class TestMakeError:

    @pytest.mark.parametrize("noise_level", [0.0, 1e-3, 1e-1, 1.0])
    def test_inner_called(self, sf3F: Surface, noise_level: float):
        with mock.patch(
            "localuf.noise.main._Uniform.make_error",
            return_value=set(sf3F.EDGES),
        ) as m:
            error = sf3F.make_error(noise_level)
            m.assert_called_once_with(noise_level)
        assert error == set(sf3F.EDGES)

    def test_exclude_future_boundary_forward(self, rp3_forward: Repetition):
        self._exclude_temproal_boundary_edges_helper(rp3_forward)

    def test_exclude_future_boundary_frugal(self, rp3_frugal: Repetition):
        self._exclude_temproal_boundary_edges_helper(rp3_frugal)

    def _exclude_temproal_boundary_edges_helper(self, repetition: Repetition):
        NOISE_LEVEL = 1
        h = repetition.SCHEME.WINDOW_HEIGHT
        future_boundary_edges = {((j, h-1), (j, h)) for j in range(0, repetition.D-1)}
        with mock.patch(
            "localuf.noise.main._Uniform.make_error",
            return_value=future_boundary_edges,
        ) as m:
            error = repetition.make_error(NOISE_LEVEL, exclude_future_boundary=True)
            m.assert_called_once_with(NOISE_LEVEL)
        assert error == set()

    def test_exclude_future_boundary_surface(self, sfCL_OL: Surface):
        NOISE_LEVEL = 1
        h = sfCL_OL.SCHEME.WINDOW_HEIGHT
        future_boundary_edges = {e for e in sfCL_OL.EDGES if any(
            v[sfCL_OL.TIME_AXIS] == h for v in e)}
        with mock.patch(
            "localuf.noise.main.CircuitLevel.make_error",
            return_value=future_boundary_edges,
        ) as m:
            error = sfCL_OL.make_error(NOISE_LEVEL, exclude_future_boundary=True)
            m.assert_called_once_with(NOISE_LEVEL)
        assert error == set()


def test_get_syndrome(
        sf3F: Surface,
        sf3T: Surface,
        error_2D: set[Edge],
        error_3D: set[Edge],
):
    assert sf3F.get_syndrome(set()) == set()
    assert sf3F.get_syndrome(error_2D) == {(0, 1)}
    assert sf3T.get_syndrome(error_3D) == {(0, 1, 1)}


def test_compose_errors(sf3F: Surface, error_2D: set[Edge]):
    assert sf3F.compose_errors() == set()
    assert sf3F.compose_errors(error_2D) == error_2D
    assert sf3F.compose_errors(error_2D, error_2D) == set()
    # test error_2D not modified
    assert sf3F.compose_errors(error_2D, {((0, -1), (0, 0))}) == {((0, 0), (0, 1)),}


def test_get_logical_error(sf3F: Surface):
    with mock.patch("localuf._schemes.Batch.get_logical_error") as m:
        sf3F.get_logical_error(set())
        m.assert_called_once_with(set())


def test_graph_attribute(sf: Surface):
    with mock.patch("networkx.Graph") as mock_Graph:
        sf.GRAPH
        mock_Graph.assert_called_once_with(
            sf.EDGES,
            D=sf.D,
            SCHEME=str(sf.SCHEME),
        )
    del sf.GRAPH
    g = sf.GRAPH
    assert type(g) is nx.Graph
    assert len(sf.EDGES) == len(g.edges)
    assert set(g.nodes) == set(sf.NODES)
    assert g.graph['D'] == sf.D
    assert g.graph['SCHEME'] == str(sf.SCHEME)
    # test cached property
    with mock.patch("networkx.Graph") as mock_Graph:
        sf.GRAPH
        mock_Graph.assert_not_called()


def test_draw(sf3F: Surface):
    with (
        mock.patch("localuf.Surface.get_syndrome") as mock_get_syndrome,
        mock.patch("localuf.Surface.get_pos") as mock_get_pos,
        mock.patch("localuf.Surface.get_node_color") as mock_get_node_color,
        mock.patch("networkx.draw") as mock_draw,
    ):
        sf3F.draw()
        mock_get_syndrome.assert_called_once_with(set())
        mock_get_pos.assert_called_once_with(constants.DEFAULT_X_OFFSET)
        mock_get_node_color.assert_called_once()
        mock_draw.assert_called_once()


def test_get_node_color(sf3F: Surface):
    ls = sf3F.get_node_color(
        set(),
        boundary_color='a',
        defect_color='b',
        nondefect_color='c',
    )
    assert type(ls) is list
    assert len(ls) == len(sf3F.NODES)
    assert set(ls) == {'a', 'c'}
    ls = sf3F.get_node_color(
        {(0, 1)},
        boundary_color='d',
        defect_color='e',
        nondefect_color='f',
    )
    assert set(ls) == {'d', 'e', 'f'}
    ls = sf3F.get_node_color(
        {v for v in sf3F.NODES if sf3F.is_boundary(v)},
        boundary_color='g',
        defect_color='h',
        nondefect_color='i',
    )
    assert set(ls) == {'h', 'i'}
    ls = sf3F.get_node_color(
        set(),
        nodelist=[]
    )
    assert ls == []


"""Plot runtime data from ``sim.runtime``.

Available functions:

* mean
* distribution
* distributions
* violin
"""

from collections.abc import Callable
import itertools
from string import ascii_lowercase
from typing import Iterable, Literal, Sequence

from matplotlib import pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.container import ErrorbarContainer
from pandas import DataFrame, Index, Series


DEFAULT_FILL_ALPHA = 0.15


def mean(
        data: DataFrame,
        title: str = '',
        per_measurement_round: bool = False,
        layers_per_sample: Callable[[int], int] = lambda d: d,
        yerr_shows: Literal['sem', 'std'] = 'sem',
        plot_noise_levels: Sequence[float] | None = None,
        legend: None | bool = None,
        grid: bool = False,
        xlabel: None | str = None,
        ylabel: None | str = None,
        base_color: None | tuple[float, float, float] | str = None,
        fill_between: bool = True,
        fill_alpha: float = DEFAULT_FILL_ALPHA,
        capsize: float = 2,
        quantile: None | float = None,
        quantile_linestyle: str = '--',
        **kwargs,
):
    """Plot mean timestep count against code distance.
    
    
    :param data: a DataFrame where each
        column a (distance, probability);
        row, a runtime sample.
    :param title: plot title.
    :param per_measurement_round: whether to divide runtime by measurement round count.
    :param layers_per_sample: a function with input ``d`` that outputs
        the measurement round count per row of ``data``.
        Affects output only if ``per_measurement_round``.
    :param yerr_shows: what errorbars show:
        either ``'sem'`` for standard error,
        or ``'std'`` for standard deviation.
    :param plot_noise_levels: sequence specifying which noise levels to plot, in case want to omit any.
    :param base_color: a single color for all errorbars and their connecting lines.
        Decreasing noise level is then shown by increasing opacity.
        If ``None``, each noise level is shown by a different, fully opaque color.
    :param fill_between: whether to use ``fill_between`` instead of ``errorbar``.
    :param fill_alpha: alpha value for the filled area.
    :param capsize: length of error bar caps in points.
    :param quantile: optional quantile (in the interval [0, 1]) to line plot.
    :param quantile_linestyle: linestyle for the quantile line.
    :param kwargs: passed to either ``errorbar`` or ``fill_between``
        depending on which is used.
    
    
    :return data_copy: A copy of ``data`` with runtimes divided by distance
        if ``per_measurement_round`` else an exact deep copy of ``data``.
    :return containers: A dictionary where each key is a noise level;
        value, the ``ErrorbarContainer`` for that noise level.
    """
    containers: dict[float, ErrorbarContainer | list[Line2D]] = {}
    ds: Index[int] = data.columns.get_level_values('d').unique()
    data_copy = data.copy()
    if per_measurement_round:
        for d in ds:
            data_copy[d] = data[d] / layers_per_sample(d)
        default_ylabel = 'mean runtime per measurement round'
    else:
        default_ylabel = 'mean runtime'
    noise_levels: Sequence[float] = data.columns.get_level_values('p').unique() \
        if plot_noise_levels is None else plot_noise_levels # type: ignore
    if base_color is None:
        colors = itertools.repeat(None)
        if legend is None:
            legend = True
    else:
        if legend is None:
            legend = False
        p_count = len(noise_levels)
        colors = [(base_color, k/p_count) for k in range(p_count, 0, -1)]
    for noise_level, color in zip(noise_levels, colors):
        df: DataFrame = data_copy.xs(noise_level, level='p', axis=1) # type: ignore

        if yerr_shows == 'sem':
            yerr: Series[float] = df.sem()
        elif yerr_shows == 'std':
            yerr: Series[float] = df.std()
        else:
            raise ValueError(f'invalid yerr_shows: {yerr_shows}')

        if fill_between:
            container = plt.plot(
                ds,
                df.mean(),
                label=f'{noise_level:.1e}',
                color=color,
            )
            lo = df.mean() - yerr
            hi = df.mean() + yerr
            plt.fill_between(
                x=ds,
                y1=lo,
                y2=hi,
                color=container[0].get_color(),
                alpha=fill_alpha,
                linewidth=0,
                **kwargs,
            )
        else:
            container = plt.errorbar(
                x=ds,
                y=df.mean(),
                yerr=yerr,
                capsize=capsize,
                label=f'{noise_level:.1e}',
                color=color,
                **kwargs,
            )
        if quantile is not None:
            plt.plot(
                ds,
                df.quantile(quantile),
                color=container[0].get_color(),
                linestyle=quantile_linestyle,
            )
        containers[noise_level] = container
    plt.xticks(ds);
    if legend: plt.legend(title=r'$p =\dots$', reverse=True)
    if grid: plt.grid(which='both')
    if xlabel is None: xlabel = 'code distance'
    if ylabel is None: ylabel = default_ylabel
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    if title: plt.title(title)
    return data_copy, containers


def distribution(
        data: DataFrame,
        noise_level: float,
        bins=80,
        horizontal=True,
        figsize: None | tuple[float, float] = None,
        log_scale=True,
        grid=False,
        global_range=True,
):
    df: DataFrame = data.xs(noise_level, level='p', axis=1) # type: ignore
    ds = df.columns
    min_runtime = min(df.min())
    max_runtime = max(df.max())
    range_ = (min_runtime, max_runtime) if global_range else None
    if horizontal:
        if figsize is None:
            figsize=(15, 4)
        f = plt.figure(figsize=figsize)
        for k, d in enumerate(ds, start=1):
            ax = plt.subplot(1, len(ds), k)
            if bins:
                df[d].hist(
                    bins=bins,
                    range=range_,
                    orientation='horizontal',
                )
            else:
                vc = df[d].value_counts(); plt.barh(vc.index, vc, height=1)
            plt.grid(grid)
            plt.ylim(0, max_runtime)
            if k == 1:
                plt.ylabel('timestep count')
            else:
                ax.set_yticklabels([])
            if log_scale:
                plt.xscale('log')
            plt.xlabel(f'$d = {d}$')
            plt.axhline(y=df[d].mean(), color='magenta')
            plt.axhline(y=df[d].max(), color='r')
    else:
        if figsize is None:
            figsize=(5, 10)
        f = plt.figure(figsize=figsize)
        for k, d in enumerate(ds, start=1):
            ax = plt.subplot(len(ds), 1, k)
            if bins:
                df[d].hist(
                    bins=bins,
                    range=range_,
                )
            else:
                vc = df[d].value_counts(); plt.bar(vc.index, vc, width=1)
            plt.grid(grid)
            plt.xlim(0, max_runtime)
            if k == len(ds):
                plt.xlabel('timestep count')
            else:
                ax.set_xticklabels([])
            if log_scale:
                plt.yscale('log')
            plt.ylabel(f'$d = {d}$')
            plt.axvline(x=df[d].mean(), color='magenta')
            plt.axvline(x=df[d].max(), color='r')
    return f


def distributions(
        data: Sequence[DataFrame],
        noise_level: float,
        bins: int | Iterable[int] = 80,
        figsize: None | tuple[float, float] = None,
        log_scale=True,
        grid=False,
        global_range=True,
        show_xticks=False,
        subplots_hspace=0.05,
        quantile: float = 1,
        supxlabel_y: float = 0.15,
        **kwargs_for_ylabel,
):
    """Histogram runtime distributions for each DataFrame in ``data``.
    
    
    :param data: sequence of DataFrames. In each DataFrame, each
        column a (distance, probability);
        row, a runtime sample.
    :param noise_level: noise level associated to the runtimes histogrammed.
    :param bins: bin count in each histogram.
        If an int, use same bin count for all entries in ``data``.
        If any bin count is 0, set bin width to 1.
    :param global_range: whether to use same bins for all distances within a DataFrame.
    :param quantile: the quantile (in the interval [0, 1]) to draw as a horizontal red line.
        Default is 1 i.e. the maximum of the sample.
    :param supxlabel_y: y-coordinate for the figure x-label.
    :param kwargs_for_ylabel: passed to ``plt.ylabel`` for the leftmost subplot in each row.
    """
    if isinstance(bins, int):
        bins = itertools.repeat(bins)
    ds: Index[int] = next(iter(data)).xs(noise_level, level='p', axis=1).columns # type: ignore
    if figsize is None:
        figsize=(len(ds), 3)
    w, h = figsize
    figsize = (w, h*len(data))
    f = plt.figure(figsize=figsize)
    for i, (dfmi, bin_count, letter) in enumerate(zip(data, bins, ascii_lowercase)):
        df: DataFrame = dfmi.xs(noise_level, level='p', axis=1) # type: ignore
        min_runtime = min(df.min())
        max_runtime = max(df.max())
        range_ = (min_runtime, max_runtime) if global_range else None
        for j, d in enumerate(ds, start=1):
            ax = plt.subplot(len(data), len(ds), i*len(ds) + j)
            if bin_count:
                df[d].hist(
                    bins=bin_count,
                    range=range_,
                    orientation='horizontal',
                )
            else:
                vc = df[d].value_counts(); plt.barh(vc.index, vc, height=1)
            plt.grid(grid)
            plt.ylim(0, max_runtime)
            if j == 1:
                plt.ylabel(f'({letter})', **kwargs_for_ylabel)
            else:
                ax.set_yticklabels([])
            if log_scale:
                plt.xscale('log')
            if i == len(data)-1:
                plt.xlabel(str(d))
            plt.axhline(y=df[d].mean(), color='magenta')
            plt.axhline(y=df[d].quantile(quantile), color='r')
    f.supxlabel('code distance $d$', size='medium', y=supxlabel_y)
    f.tight_layout()
    if not show_xticks:
        for ax in f.axes:
            ax.set_xticks([])
            ax.set_xticks([], minor=True)
        f.subplots_adjust(hspace=subplots_hspace, wspace=0)
    return f


def violin(
        data: DataFrame,
        noise_level: float,
        title='',
        widths=1,
        showextrema=False,
        capsize: float = 3,
        errorbar_kwargs: None | dict = None,
        **kwargs_for_violinplot,
):
    """Violin plot of runtime distributions for a given noise level.
    
    
    :param capsize: length of error bar caps in points.
    """
    df: DataFrame = data.xs(noise_level, level='p', axis=1) # type: ignore
    if errorbar_kwargs is None: errorbar_kwargs = {}
    parts = plt.violinplot(
        df,
        positions=list(df.columns),
        widths=widths,
        showextrema=showextrema,
        **kwargs_for_violinplot,
    )
    for pc in parts['bodies']: # type: ignore
        pc.set_alpha(1)
    plt.errorbar(
        x=df.columns,
        y=df.mean(),
        yerr=df.std(),
        fmt='.',
        capsize=capsize,
        linestyle='none',
        **errorbar_kwargs,
    )
    plt.xticks(df.columns)
    plt.grid(which='both')
    plt.xlabel('code distance')
    plt.ylabel('timestep count')
    if title:
        plt.title(title)


from localuf.type_aliases import Node, Edge
from localuf.decoders import uf
from localuf.constants import Growth

class NodeUF(uf.UF):
    """The graph used by original UF decoder.
    
    Extends ``UF`` class.
    Overriden attributes:
    
    * ``clusters`` and ``active_clusters`` now use ``_NodeCluster`` rather than ``_Cluster``.
    
    Overriden methods:
    
    * ``reset``.
    * ``_growth_round``.
    * ``_grow``.
    * ``_union``.
    """

    def reset(self):
        super().reset()
        self.clusters = {v: _NodeCluster(self, v) for v in self.CODE.NODES}
        self.active_clusters: set[_NodeCluster] = set()

    def _growth_round(
            self,
            log_history,
            clusters_to_grow: 'set[_NodeCluster] | None' = None,
    ):
        if clusters_to_grow is None:
            clusters_to_grow = self.active_clusters
        merge_ls: list[Edge] = []
        changed_edges: set[Edge] = set()
        for cluster in clusters_to_grow:
            self._grow(cluster, merge_ls, changed_edges)
        for u, v in merge_ls:
            cu = self.clusters[self._find(u)]
            cv = self.clusters[self._find(v)]
            self._FORESTER.merge(cu, cv, (u, v)) # type: ignore
        if log_history:
            self.changed_edges = changed_edges
            self.append_history()

    def _grow(
            self,
            cluster: '_NodeCluster',
            merge_ls,
            changed_edges,
    ):
        ex_frontier: set[Node] = set()
        for v in cluster.frontier:

            edges_to_grow = {e for e in self.CODE.INCIDENT_EDGES[v]
                if self.growth[e] in self._ACTIVE_GROWTH_VALUES}
            changed_edges.update(edges_to_grow)
            # if node no longer on cluster frontier, mark to remove
            if not edges_to_grow:
                ex_frontier.add(v)

            for e in edges_to_grow:
                self.growth[e] += Growth.INCREMENT
                if self.growth[e] is Growth.FULL:
                    merge_ls.append(e)

        # remove nodes that are no longer on the cluster frontier
        cluster.frontier.difference_update(ex_frontier)

    def _union(self, cu: '_NodeCluster', cv: '_NodeCluster'):
        if cu.size >= cv.size:
            larger, smaller = cu, cv
        else:
            larger, smaller = cv, cu
        self.parents[smaller.root] = larger.root
        old_larger_size = larger.size
        larger.size += smaller.size
        larger.odd ^= smaller.odd  # logical XOR
        larger.frontier.update(smaller.frontier)
        self._INCLINATION.update_boundary(larger, smaller)
        self._update_self_after_union(larger, smaller, old_larger_size) # type: ignore


class _NodeCluster(uf.BaseCluster):
    """A UF cluster that tracks the nodes on its frontier.
    
    Extends ``BaseCluster``.
    
    Additional attributes:
    
    * ``frontier`` the detectors in the cluster that are incident to >=1 active edge.
    """

    def __init__(self, uf: NodeUF, root: Node):
        super().__init__(uf, root)
        self.frontier: set[Node] = set() if uf.CODE.is_boundary(root) else {root}


from unittest import mock

import pytest

from localuf.constants import Growth
from localuf.codes import Surface
from localuf.decoders.uf import UF, _Cluster


@pytest.fixture
def dynamic_uf7F(sf7F: Surface):
    decoder = UF(sf7F, dynamic=True)
    decoder.history = []
    return decoder


@pytest.fixture
def first_changed_edges():
    return {
        ((0, -1), (0, 0)),
        ((0, 0), (0, 1)),
        ((0, 0), (1, 0)),
        ((0, 1), (0, 2)),
        ((0, 1), (1, 1)),
        ((0, 3), (0, 4)),
        ((0, 4), (0, 5)),
        ((0, 4), (1, 4)),
        ((1, 0), (2, 0)),
        ((1, 3), (1, 4)),
        ((1, 4), (1, 5)),
        ((1, 4), (2, 4)),
        ((2, -1), (2, 0)),
        ((2, 0), (2, 1)),
        ((2, 0), (3, 0)),
        ((2, 4), (3, 4)),
        ((3, 3), (3, 4)),
        ((3, 3), (4, 3)),
        ((3, 4), (3, 5)),
        ((3, 4), (4, 4)),
        ((4, 2), (4, 3)),
        ((4, 3), (4, 4)),
        ((4, 3), (5, 3)),
        ((4, 4), (4, 5)),
        ((4, 4), (5, 4))
    }

@pytest.fixture
def uf_after_union(
    uf7F: UF,
    get_uf_after_union,
    syndrome7F
) -> tuple[UF, _Cluster, _Cluster]:
    uf7F.load(syndrome7F)
    return get_uf_after_union(uf7F)


def test_reset(uf7F: UF):
    uf7F.syndrome = {(0, 0)}
    uf7F.parents = {}
    uf7F.clusters = {}
    del uf7F.active_clusters
    uf7F.changed_edges = set()
    uf7F.forest = []
    with mock.patch('localuf.decoders.uf.BaseUF.reset') as mock_reset:
        uf7F.reset()
        mock_reset.assert_called_once_with()
    assert uf7F.syndrome == set()
    nodes = uf7F.CODE.NODES
    assert uf7F.parents == dict(zip(nodes, nodes))
    assert set(uf7F.clusters.keys()) == set(nodes)
    assert uf7F.active_clusters == set()
    assert not hasattr(uf7F, 'changed_edges')
    assert not hasattr(uf7F, 'forest')


def test_validate_static(uf7F: UF, syndrome7F):
    uf7F.validate(syndrome7F)
    assert len(uf7F.clusters) == 26
    assert uf7F.active_clusters == set()
    # test erasure
    covered_nodes = set()
    for e in uf7F.erasure:
        covered_nodes.update(e)
    assert uf7F.syndrome.issubset(covered_nodes)
    assert len(uf7F.erasure) == 38

def test_validate_dynamic(dynamic_uf7F: UF, syndrome7F):
    dynamic_uf7F.validate(syndrome7F)
    assert len(dynamic_uf7F.clusters) == 27
    assert dynamic_uf7F.active_clusters == set()
    # test erasure
    covered_nodes = set()
    for e in dynamic_uf7F.erasure:
        covered_nodes.update(e)
    assert dynamic_uf7F.syndrome.issubset(covered_nodes)
    assert len(dynamic_uf7F.erasure) == 29

def test_growth_round_static(
        uf7F: UF,
        syndrome7F,
        first_changed_edges,
):
    uf7F.load(syndrome7F)
    log_history = True
    uf7F._growth_round(log_history)
    assert uf7F.history[0].changed_edges == first_changed_edges
    c = uf7F.clusters[uf7F._find((0, 0))]
    # test inactive edges removed
    assert c.vision == {
        ((0, -1), (0, 0)),
        ((0, 1), (0, 2)),
        ((0, 0), (1, 0)),
        ((0, 1), (1, 1))
    }
    while uf7F.active_clusters:
        assert all(val in {
            Growth.UNGROWN,
            Growth.HALF,
            Growth.FULL
        } for val in uf7F.growth.values())
        uf7F._growth_round(log_history)

def test_growth_round_dynamic(
        dynamic_uf7F: UF,
        syndrome7F,
        first_changed_edges,
):
    dynamic_uf7F.load(syndrome7F)
    log_history = True
    dynamic_uf7F._growth_round(log_history)
    assert dynamic_uf7F.history[0].changed_edges == first_changed_edges
    c = dynamic_uf7F.clusters[dynamic_uf7F._find((0, 0))]
    # test inactive edges removed
    assert c.vision == {
        ((0, -1), (0, 0)),
        ((0, 1), (0, 2)),
        ((0, 0), (1, 0)),
        ((0, 1), (1, 1))
    }
    while dynamic_uf7F.active_clusters:
        assert all(v in {
            Growth.BURNT,
            Growth.UNGROWN,
            Growth.HALF,
            Growth.FULL
        } for v in dynamic_uf7F.growth.values())
        dynamic_uf7F._growth_round(log_history)

def test_grow(uf7F: UF, fixture_test_grow):
    fixture_test_grow(uf7F)

def test_find(uf7F: UF, uvw):
    # setup
    u, v, w = uvw
    uf7F.parents[u] = v
    uf7F.parents[v] = w
    # test output
    assert uf7F._find(u) == w
    # test path compression
    assert uf7F.parents[u] == w

def test_union(uf_after_union, fixture_test_union):
    fixture_test_union(uf_after_union)

def test_update_self_after_union(uf_after_union):
    uf, larger, smaller = uf_after_union
    # check smaller cluster deleted
    assert smaller.root not in uf.clusters
    assert smaller not in uf.active_clusters
    # check larger cluster in clusters but not in active_clusters
    assert larger.root in uf.clusters
    assert larger not in uf.active_clusters

def test_peel(uf7F: UF, syndrome7F):
    uf7F.validate(syndrome7F)
    uf7F.peel()
    syndrome_from_correction = uf7F.CODE.get_syndrome(uf7F.correction)
    assert syndrome_from_correction == uf7F.syndrome

def test_make_forest(dynamic_uf7F: UF, syndrome7F):
    dynamic_uf7F.validate(syndrome7F)
    forest = dynamic_uf7F._make_forest()
    assert len(forest) == len(dynamic_uf7F.erasure)

def test_make_tree(uf7F: UF, syndrome7F):
    uf7F.validate(syndrome7F)
    assert uf7F._make_tree((0, 2)) == []
    assert uf7F._make_tree((0, 0)) == [(((0, 0), (0, 1)), (0, 1))]
    first, *rest = uf7F._make_tree((2, -1))
    assert first == (((2, -1), (2, 0)), (2, 0))
    assert set(rest) == {
        (((1, 0), (2, 0)), (1, 0)),
        (((2, 0), (2, 1)), (2, 1)),
        (((2, 0), (3, 0)), (3, 0))
    }
    n_nodes_in_big_cluster = 26
    assert len(uf7F._make_tree((4, 6))) == n_nodes_in_big_cluster - 1


def test_weigh_correction(uf7F: UF):
    flip_probability = 0.5
    with mock.patch('localuf.noise.main._Uniform.get_edge_weights') as mock_get_edge_weights:
        mock_get_edge_weights.return_value = {
            ((0, 0), (0, 1)): (flip_probability, 1),
            ((0, 1), (0, 2)): (flip_probability, 2),
            ((0, 2), (0, 3)): (flip_probability, 3),
            ((0, 3), (0, 4)): (flip_probability, 4),
        }
        weight = uf7F._weigh_correction()
        mock_get_edge_weights.assert_called_once_with(None)
        assert weight == 0

        uf7F.correction = {
            ((0, 0), (0, 1)),
            ((0, 1), (0, 2)),
            ((0, 2), (0, 3)),
            ((0, 3), (0, 4)),
        }
        weight = uf7F._weigh_correction()
        assert weight == 10


def test_complementary_gap(uf7F: UF):
    uf7F.correction = set()
    with pytest.raises(ValueError):
        uf7F.complementary_gap()


from unittest import mock

import pytest

from localuf.decoders.uf import _Cluster
from localuf.decoders import NodeBUF

@pytest.fixture
def node_buf_after_union(
    node_buf: NodeBUF,
    get_uf_after_union,
    syndrome7F,
) -> tuple[NodeBUF, _Cluster, _Cluster]:
    node_buf.load(syndrome7F)
    return get_uf_after_union(node_buf)

def test_reset(node_buf: NodeBUF):
    with mock.patch("localuf.decoders.node_uf.NodeUF.reset") as super_reset:
        node_buf.reset()
        super_reset.assert_called_once_with()
    # test correct number of buckets
    assert len(node_buf.buckets) == len(node_buf.CODE.NODES)
    # test all buckets empty
    assert all(bucket == set() for bucket in node_buf.buckets.values())

def test_load(node_buf, syndrome7F):
    with mock.patch("localuf.decoders.node_uf.NodeUF.load") as super_load:
        node_buf.load(syndrome7F)
        super_load.assert_called_once_with(syndrome7F)
    # test bucket 1 initially has all defects
    assert {cluster.root for cluster in node_buf.buckets[1]} == node_buf.syndrome
    # test active_clusters attribute replaced by buckets
    assert not hasattr(node_buf, 'active_clusters')

def test_validate(node_buf, syndrome7F):

    bucket = {_Cluster(node_buf, (0, 0))}
    node_buf.buckets = {1: bucket}

    def fake_growth_round(log_history, clusters_to_grow: set[_Cluster]):
        clusters_to_grow.clear()

    with (
        mock.patch("localuf.decoders.node_buf.NodeBUF.load") as mock_load,
        mock.patch(
            "localuf.decoders.node_buf.NodeBUF._growth_round",
            side_effect=fake_growth_round,
        ) as mock_growth_round,
    ):
        node_buf.validate(syndrome7F)
        mock_load.assert_called_once_with(syndrome7F)
        mock_growth_round.assert_called_once_with(
            False,
            clusters_to_grow=bucket,
        )

def test_update_self_after_union(
        node_buf_after_union,
        fixture_test_update_self_after_union,
):
    fixture_test_update_self_after_union(node_buf_after_union)